import os
import logging
from twisted.internet.defer import inlineCallbacks, returnValue

from landscape.lib.fetch import fetch_async
from landscape.lib.fs import read_file
from landscape.lib.lsb_release import LSB_RELEASE_FILENAME, parse_lsb_release
from landscape.lib.cloud import fetch_ec2_meta_data
from landscape.lib.network import get_fqdn
from landscape.monitor.plugin import MonitorPlugin

METADATA_RETRY_MAX = 3  # Number of retries to get EC2 meta-data


class DistributionInfoError(Exception):
    pass


class ComputerInfo(MonitorPlugin):
    """Plugin captures and reports basic computer information."""

    persist_name = "computer-info"
    scope = "computer"

    def __init__(self, get_fqdn=get_fqdn,
                 meminfo_filename="/proc/meminfo",
                 lsb_release_filename=LSB_RELEASE_FILENAME,
                 root_path="/", fetch_async=fetch_async):
        self._get_fqdn = get_fqdn
        self._meminfo_filename = meminfo_filename
        self._lsb_release_filename = lsb_release_filename
        self._root_path = root_path
        self._cloud_instance_metadata = None
        self._cloud_retries = 0
        self._fetch_async = fetch_async

    def register(self, registry):
        super(ComputerInfo, self).register(registry)
        self._annotations_path = registry.config.annotations_path
        self.call_on_accepted("computer-info",
                              self.send_computer_message, True)
        self.call_on_accepted("distribution-info",
                              self.send_distribution_message, True)
        self.call_on_accepted("cloud-instance-metadata",
                              self.send_cloud_instance_metadata_message, True)

    def send_computer_message(self, urgent=False):
        message = self._create_computer_info_message()
        if message:
            message["type"] = "computer-info"
            logging.info("Queueing message with updated computer info.")
            self.registry.broker.send_message(message, self._session_id,
                                              urgent=urgent)

    def send_distribution_message(self, urgent=False):
        message = self._create_distribution_info_message()
        if message:
            message["type"] = "distribution-info"
            logging.info("Queueing message with updated distribution info.")
            self.registry.broker.send_message(message, self._session_id,
                                              urgent=urgent)

    @inlineCallbacks
    def send_cloud_instance_metadata_message(self, urgent=False):
        message = yield self._create_cloud_instance_metadata_message()
        if message:
            message["type"] = "cloud-instance-metadata"
            logging.info("Queueing message with updated cloud instance "
                         "metadata.")
            self.registry.broker.send_message(message, self._session_id,
                                              urgent=urgent)

    def exchange(self, urgent=False):
        broker = self.registry.broker
        broker.call_if_accepted("computer-info",
                                self.send_computer_message, urgent)
        broker.call_if_accepted("distribution-info",
                                self.send_distribution_message, urgent)
        broker.call_if_accepted("cloud-instance-metadata",
                                self.send_cloud_instance_metadata_message,
                                urgent)

    def _create_computer_info_message(self):
        message = {}
        self._add_if_new(message, "hostname", self._get_fqdn())
        total_memory, total_swap = self._get_memory_info()
        self._add_if_new(message, "total-memory", total_memory)
        self._add_if_new(message, "total-swap", total_swap)
        annotations = {}
        if os.path.exists(self._annotations_path):
            for key in os.listdir(self._annotations_path):
                annotations[key] = read_file(
                    os.path.join(self._annotations_path, key))

        if annotations:
            self._add_if_new(message, "annotations", annotations)
        return message

    def _add_if_new(self, message, key, value):
        if value != self._persist.get(key):
            self._persist.set(key, value)
            message[key] = value

    def _create_distribution_info_message(self):
        message = self._get_distribution_info()
        if message != self._persist.get("distribution-info"):
            self._persist.set("distribution-info", message)
            return message
        return None

    def _get_memory_info(self):
        """Get details in megabytes and return a C{(memory, swap)} tuple."""
        message = {}
        file = open(self._meminfo_filename)
        for line in file:
            if line != '\n':
                parts = line.split(":")
                key = parts[0]
                if key in ["MemTotal", "SwapTotal"]:
                    value = int(parts[1].strip().split(" ")[0])
                    message[key] = value
        file.close()
        return (message["MemTotal"] // 1024, message["SwapTotal"] // 1024)

    def _get_distribution_info(self):
        """Get details about the distribution."""
        message = {}
        message.update(parse_lsb_release(self._lsb_release_filename))
        return message

    @inlineCallbacks
    def _create_cloud_instance_metadata_message(self):
        """Fetch cloud metadata and insert it in a message."""
        message = None
        if (self._cloud_instance_metadata is None and
            self._cloud_retries < METADATA_RETRY_MAX):
            self._cloud_instance_metadata = yield self._fetch_ec2_meta_data()
            message = self._cloud_instance_metadata
        returnValue(message)

    def _fetch_ec2_meta_data(self):
        """Fetch information about the cloud instance."""
        if self._cloud_retries == 0:
            logging.info("Querying cloud meta-data.")
        deferred = fetch_ec2_meta_data(self._fetch_async)

        def log_no_meta_data_found(error):
            self._cloud_retries += 1
            if self._cloud_retries >= METADATA_RETRY_MAX:
                logging.info("No cloud meta-data available. %s" %
                        error.getErrorMessage())

        def log_success(result):
            logging.info("Acquired cloud meta-data.")
            return result

        deferred.addCallback(log_success)
        deferred.addErrback(log_no_meta_data_found)
        return deferred



from landscape.tests.helpers import LandscapeTest
from landscape.patch import UpgradeManager

from landscape.upgraders import broker


class TestBrokerUpgraders(LandscapeTest):

    def test_broker_upgrade_manager(self):
        self.assertEqual(type(broker.upgrade_manager), UpgradeManager)



"""Deployment code for the monitor."""

import os

from landscape.service import LandscapeService, run_landscape_service
from landscape.amp import ComponentPublisher
from landscape.broker.registration import RegistrationHandler, Identity
from landscape.broker.config import BrokerConfiguration
from landscape.broker.transport import HTTPTransport
from landscape.broker.exchange import MessageExchange
from landscape.broker.exchangestore import ExchangeStore
from landscape.broker.ping import Pinger
from landscape.broker.store import get_default_message_store
from landscape.broker.server import BrokerServer


class BrokerService(LandscapeService):
    """The core C{Service} of the Landscape Broker C{Application}.

    The Landscape broker service handles all the communication between the
    client and server.  When started it creates and runs all necessary
    components to exchange messages with the Landscape server.

    @cvar service_name: C{broker}

    @ivar persist_filename: Path to broker-specific persistent data.
    @ivar persist: A L{Persist} object saving and loading data from
        C{self.persist_filename}.
    @ivar message_store: A L{MessageStore} used by the C{exchanger} to
        queue outgoing messages.
    @ivar transport: An L{HTTPTransport} used by the C{exchanger} to deliver
        messages.
    @ivar identity: The L{Identity} of the Landscape client the broker runs on.
    @ivar exchanger: The L{MessageExchange} exchanges messages with the server.
    @ivar pinger: The L{Pinger} checks if the server has new messages for us.
    @ivar registration: The L{RegistrationHandler} performs the initial
        registration.

    @param config: A L{BrokerConfiguration}.
    """

    transport_factory = HTTPTransport
    pinger_factory = Pinger
    service_name = BrokerServer.name

    def __init__(self, config):
        self.persist_filename = os.path.join(
            config.data_path, "%s.bpickle" % (self.service_name,))
        super(BrokerService, self).__init__(config)

        self.transport = self.transport_factory(
            self.reactor, config.url, config.ssl_public_key)
        self.message_store = get_default_message_store(
            self.persist, config.message_store_path)
        self.identity = Identity(self.config, self.persist)
        exchange_store = ExchangeStore(self.config.exchange_store_path)
        self.exchanger = MessageExchange(
            self.reactor, self.message_store, self.transport, self.identity,
            exchange_store, config)
        self.pinger = self.pinger_factory(
            self.reactor, self.identity, self.exchanger, config)
        self.registration = RegistrationHandler(
            config, self.identity, self.reactor, self.exchanger, self.pinger,
            self.message_store)
        self.broker = BrokerServer(self.config, self.reactor, self.exchanger,
                                   self.registration, self.message_store,
                                   self.pinger)
        self.publisher = ComponentPublisher(self.broker, self.reactor,
                                            self.config)

    def startService(self):
        """Start the broker.

        Create a L{BrokerServer} listening on C{broker_socket_path} for clients
        connecting with the L{BrokerServerConnector}, and start the
        L{MessageExchange} and L{Pinger} services.
        """
        super(BrokerService, self).startService()
        self.publisher.start()
        self.exchanger.start()
        self.pinger.start()

    def stopService(self):
        """Stop the broker."""
        self.publisher.stop()
        self.exchanger.stop()
        self.pinger.stop()
        super(BrokerService, self).stopService()


def run(args):
    """Run the application, given some command line arguments."""
    run_landscape_service(BrokerConfiguration, BrokerService, args)



from twisted.internet.error import ConnectError
from twisted.internet.task import Clock

from landscape.tests.helpers import LandscapeTest
from landscape.reactor import FakeReactor
from landscape.deployment import Configuration
from landscape.amp import ComponentPublisher, ComponentConnector, remote
from landscape.lib.amp import MethodCallError


class TestComponent(object):

    name = "test"

    @remote
    def ping(self):
        return True

    def non_remote(self):
        return False


class TestComponentConnector(ComponentConnector):

    component = TestComponent


class FakeAMP(object):

    def __init__(self, locator):
        self._locator = locator


class ComponentPublisherTest(LandscapeTest):

    def setUp(self):
        super(ComponentPublisherTest, self).setUp()
        reactor = FakeReactor()
        config = Configuration()
        config.data_path = self.makeDir()
        self.makeDir(path=config.sockets_path)
        self.component = TestComponent()
        self.publisher = ComponentPublisher(self.component, reactor, config)
        self.publisher.start()

        self.connector = TestComponentConnector(reactor, config)
        connected = self.connector.connect()
        connected.addCallback(lambda remote: setattr(self, "remote", remote))
        return connected

    def tearDown(self):
        self.connector.disconnect()
        self.publisher.stop()
        super(ComponentPublisherTest, self).tearDown()

    def test_remote_methods(self):
        """Methods decorated with @remote are accessible remotely."""
        result = self.remote.ping()
        return self.assertSuccess(result, True)

    def test_protect_non_remote(self):
        """Methods not decorated with @remote are not accessible remotely."""
        result = self.remote.non_remote()
        failure = self.failureResultOf(result)
        self.assertTrue(failure.check(MethodCallError))


class ComponentConnectorTest(LandscapeTest):

    def setUp(self):
        super(ComponentConnectorTest, self).setUp()
        self.reactor = FakeReactor()
        # XXX this should be dropped once the FakeReactor doesn't use the
        # real reactor anymore under the hood.
        self.reactor._reactor = Clock()
        self.config = Configuration()
        self.config.data_path = self.makeDir()
        self.makeDir(path=self.config.sockets_path)
        self.connector = TestComponentConnector(self.reactor, self.config)

    def test_connect_with_max_retries(self):
        """
        If C{max_retries} is passed to L{RemoteObjectConnector.connect},
        then it will give up trying to connect after that amount of times.
        """
        self.log_helper.ignore_errors("Error while connecting to test")
        deferred = self.connector.connect(max_retries=2)
        self.assertNoResult(deferred)
        return
        self.failureResultOf(deferred).trap(ConnectError)

    def test_connect_logs_errors(self):
        """
        Connection errors are logged.
        """
        self.log_helper.ignore_errors("Error while connecting to test")

        def assert_log(ignored):
            self.assertIn("Error while connecting to test",
                          self.logfile.getvalue())

        result = self.connector.connect(max_retries=0)
        self.assertFailure(result, ConnectError)
        return result.addCallback(assert_log)

    def test_connect_with_quiet(self):
        """
        If the C{quiet} option is passed, no errors will be logged.
        """
        result = self.connector.connect(max_retries=0, quiet=True)
        return self.assertFailure(result, ConnectError)

    def test_reconnect_fires_event(self):
        """
        An event is fired whenever the connection is established again after
        it has been lost.
        """
        reconnects = []
        self.reactor.call_on("test-reconnect", lambda: reconnects.append(True))

        component = TestComponent()
        publisher = ComponentPublisher(component, self.reactor, self.config)
        publisher.start()
        deferred = self.connector.connect()
        self.successResultOf(deferred)
        self.connector._connector.disconnect()  # Simulate a disconnection
        self.assertEqual([], reconnects)
        self.reactor._reactor.advance(10)
        self.assertEqual([True], reconnects)

    def test_connect_with_factor(self):
        """
        If C{factor} is passed to the L{ComponentConnector.connect} method,
        then the associated protocol factory will be set to that value.
        """
        component = TestComponent()
        publisher = ComponentPublisher(component, self.reactor, self.config)
        publisher.start()
        deferred = self.connector.connect(factor=1.0)
        remote = self.successResultOf(deferred)
        self.assertEqual(1.0, remote._factory.factor)

    def test_disconnect(self):
        """
        It is possible to call L{ComponentConnector.disconnect} multiple times,
        even if the connection has been already closed.
        """
        component = TestComponent()
        publisher = ComponentPublisher(component, self.reactor, self.config)
        publisher.start()
        self.connector.connect()
        self.connector.disconnect()
        self.connector.disconnect()

    def test_disconnect_without_connect(self):
        """
        It is possible to call L{ComponentConnector.disconnect} even if the
        connection was never established. In that case the method is
        effectively a no-op.
        """
        self.connector.disconnect()



"""
Extend the regular Twisted reactor with event-handling features.
"""
import time
import sys
import logging
import bisect

from twisted.python.failure import Failure
from twisted.internet.error import ConnectError
from twisted.internet.threads import deferToThread

from landscape.log import format_object


class InvalidID(Exception):
    """Raised when an invalid ID is used with reactor.cancel_call()."""


class CallHookError(Exception):
    """Raised when hooking on a reactor incorrectly."""


class EventID(object):
    """Unique identifier for an event handler.

    @param event_type: Name of the event type handled by the handler.
    @param pair: Binary tuple C{(handler, priority)} holding the handler
        function and its priority.
    """

    def __init__(self, event_type, pair):
        self._event_type = event_type
        self._pair = pair


class EventHandlingReactorMixin(object):
    """Fire events identified by strings and register handlers for them.

    Note that event handlers are executed synchronously when the C{fire} method
    is called, so unit-tests can generally exercise events without needing to
    run the real Twisted reactor (except of course if the event handlers
    themselves contain asynchronous calls that need the Twisted reactor
    running).
    """

    def __init__(self):
        super(EventHandlingReactorMixin, self).__init__()
        self._event_handlers = {}

    def call_on(self, event_type, handler, priority=0):
        """Register an event handler.

        The handler will be invoked every time an event of the given type
        is fired (there's no need to re-register the handler after the
        event is fired).

        @param event_type: The name of the event type to handle.
        @param handler: The function handling the given event type.
        @param priority: The priority of the given handler function.

        @return: The L{EventID} of the registered handler.
        """
        pair = (handler, priority)

        handlers = self._event_handlers.setdefault(event_type, [])
        handlers.append(pair)
        handlers.sort(key=lambda pair: pair[1])

        return EventID(event_type, pair)

    def fire(self, event_type, *args, **kwargs):
        """Fire an event of a given type.

        Call all handlers registered for the given C{event_type}, in order
        of priority.

        @param event_type: The name of the event type to fire.
        @param args: Positional arguments to pass to the registered handlers.
        @param kwargs: Keyword arguments to pass to the registered handlers.
        """
        logging.debug("Started firing %s.", event_type)
        results = []
        # Make a copy of the handlers that are registered at this point in
        # time, so we have a stable list in case handlers are cancelled
        # dynamically by executing the handlers themselves.
        handlers = list(self._event_handlers.get(event_type, ()))
        for handler, priority in handlers:
            try:
                logging.debug("Calling %s for %s with priority %d.",
                              format_object(handler), event_type, priority)
                results.append(handler(*args, **kwargs))
            except KeyboardInterrupt:
                logging.exception("Keyboard interrupt while running event "
                                  "handler %s for event type %r with "
                                  "args %r %r.", format_object(handler),
                                  event_type, args, kwargs)
                self.stop()
                raise
            except:
                logging.exception("Error running event handler %s for "
                                  "event type %r with args %r %r.",
                                  format_object(handler), event_type,
                                  args, kwargs)
        logging.debug("Finished firing %s.", event_type)
        return results

    def cancel_call(self, id):
        """Unregister an event handler.

        @param id: the L{EventID} of the handler to unregister.
        """
        if type(id) is EventID:
            self._event_handlers[id._event_type].remove(id._pair)
        else:
            raise InvalidID("EventID instance expected, received %r" % id)


class ReactorID(object):

    def __init__(self, timeout):
        self._timeout = timeout


class LandscapeReactor(EventHandlingReactorMixin):
    """Wrap and add functionalities to the Twisted reactor.

    This is essentially a facade around the L{twisted.internet.reactor} and
    will delegate to it for mostly everything except event handling features
    which are implemented using L{EventHandlingReactorMixin}.
    """

    def __init__(self):
        from twisted.internet import reactor
        from twisted.internet.task import LoopingCall
        self._LoopingCall = LoopingCall
        self._reactor = reactor
        self._cleanup()
        self.callFromThread = reactor.callFromThread
        super(LandscapeReactor, self).__init__()

    def time(self):
        """Get current time.

        @see L{time.time}
        """
        return time.time()

    def call_later(self, *args, **kwargs):
        """Call a function later.

        Simply call C{callLater(*args, **kwargs)} and return its result.

        @see: L{twisted.internet.interfaces.IReactorTime.callLater}.

        """
        return self._reactor.callLater(*args, **kwargs)

    def call_every(self, seconds, f, *args, **kwargs):
        """Call a function repeatedly.

        Create a new L{twisted.internet.task.LoopingCall} object and
        start it.

        @return: the created C{LoopingCall} object.
        """
        lc = self._LoopingCall(f, *args, **kwargs)
        lc.start(seconds, now=False)
        return lc

    def cancel_call(self, id):
        """Cancel a scheduled function or event handler.

        @param id: The function call or handler to remove. It can be an
            L{EventID}, a L{LoopingCall} or a C{IDelayedCall}, as returned
            by L{call_on}, L{call_every} and L{call_later} respectively.
        """
        if isinstance(id, EventID):
            return EventHandlingReactorMixin.cancel_call(self, id)
        if isinstance(id, self._LoopingCall):
            return id.stop()
        if id.active():
            id.cancel()

    def call_when_running(self, f):
        """Schedule a function to be called when the reactor starts running."""
        self._reactor.callWhenRunning(f)

    def call_in_main(self, f, *args, **kwargs):
        """Cause a function to be executed by the reactor thread.

        @param f: The callable object to execute.
        @param args: The arguments to call it with.
        @param kwargs: The keyword arguments to call it with.

        @see: L{twisted.internet.interfaces.IReactorThreads.callFromThread}
        """
        self._reactor.callFromThread(f, *args, **kwargs)

    def call_in_thread(self, callback, errback, f, *args, **kwargs):
        """
        Execute a callable object in a new separate thread.

        @param callback: A function to call in case C{f} was successful, it
            will be passed the return value of C{f}.
        @param errback: A function to call in case C{f} raised an exception,
            it will be pass a C{(type, value, traceback)} tuple giving
            information about the raised exception (see L{sys.exc_info}).

        @note: Both C{callback} and C{errback} will be executed in the
            the parent thread.
        """
        def on_success(result):
            if callback:
                return callback(result)

        def on_failure(failure):
            exc_info = (failure.type, failure.value, failure.tb)
            if errback:
                errback(*exc_info)
            else:
                logging.error(exc_info[1], exc_info=exc_info)

        deferred = deferToThread(f, *args, **kwargs)
        deferred.addCallback(on_success)
        deferred.addErrback(on_failure)

    def listen_unix(self, socket, factory):
        """Start listening on a Unix socket."""
        return self._reactor.listenUNIX(socket, factory, wantPID=True)

    def connect_unix(self, socket, factory):
        """Connect to a Unix socket."""
        return self._reactor.connectUNIX(socket, factory)

    def run(self):
        """Start the reactor, a C{"run"} event will be fired."""

        self.fire("run")
        self._reactor.run()
        self.fire("stop")

    def stop(self):
        """Stop the reactor, a C{"stop"} event will be fired."""
        self._reactor.stop()
        self._cleanup()

    def _cleanup(self):
        # Since the reactor is global, we should clean it up when we
        # initialize one of our wrappers.
        for call in self._reactor.getDelayedCalls():
            if call.active():
                call.cancel()


class FakeReactorID(object):

    def __init__(self, data):
        self.active = True
        self._data = data


class FakeReactor(EventHandlingReactorMixin):
    """A fake reactor with the same API of L{LandscapeReactor}.

    This reactor emulates the asychronous interface of L{LandscapeReactor}, but
    implementing it in a synchronous way, for easier unit-testing.

    Note that the C{listen_unix} method is *not* emulated, but rather inherited
    blindly from L{UnixReactorMixin}, this means that there's no way to control
    it in a synchronous way (see the docstring of the mixin). A better approach
    would be to fake the AMP transport (i.e. fake the twisted abstractions
    around Unix sockets), and implement a fake version C{listen_unix}, but this
    hasn't been done yet.
    """
    # XXX probably this shouldn't be a class attribute, but we need client-side
    # FakeReactor instaces to be aware of listening sockets created by
    # server-side FakeReactor instances.
    _socket_paths = {}

    def __init__(self):
        super(FakeReactor, self).__init__()
        self._current_time = 0
        self._calls = []
        self.hosts = {}
        self._threaded_callbacks = []

        # XXX we need a reference to the Twisted reactor as well because
        # some tests use it
        from twisted.internet import reactor
        self._reactor = reactor

    def time(self):
        return float(self._current_time)

    def call_later(self, seconds, f, *args, **kwargs):
        scheduled_time = self._current_time + seconds
        call = (scheduled_time, f, args, kwargs)
        bisect.insort_left(self._calls, call)
        return FakeReactorID(call)

    def call_every(self, seconds, f, *args, **kwargs):

        def fake():
            # update the call so that cancellation will continue
            # working with the same ID. And do it *before* the call
            # because the call might cancel it!
            call._data = self.call_later(seconds, fake)._data
            try:
                f(*args, **kwargs)
            except:
                if call.active:
                    self.cancel_call(call)
                raise
        call = self.call_later(seconds, fake)
        return call

    def cancel_call(self, id):
        if type(id) is FakeReactorID:
            if id._data in self._calls:
                self._calls.remove(id._data)
            id.active = False
        else:
            super(FakeReactor, self).cancel_call(id)

    def call_when_running(self, f):
        # Just schedule a call that will be kicked by the run() method.
        self.call_later(0, f)

    def call_in_main(self, f, *args, **kwargs):
        """Schedule a function for execution in the main thread."""
        self._threaded_callbacks.append(lambda: f(*args, **kwargs))

    def call_in_thread(self, callback, errback, f, *args, **kwargs):
        """Emulate L{LandscapeReactor.call_in_thread} without spawning threads.

        Note that running threaded callbacks here doesn't reflect reality,
        since they're usually run while the main reactor loop is active. At
        the same time, this is convenient as it means we don't need to run
        the the real Twisted reactor with to test actions performed on
        completion of specific events (e.g. L{MessageExchange.exchange} uses
        call_in_thread to run the HTTP request in a separate thread, because
        we use libcurl which is blocking). IOW, it's easier to test things
        synchronously.
        """
        self._in_thread(callback, errback, f, args, kwargs)
        self._run_threaded_callbacks()

    def listen_unix(self, socket_path, factory):

        class FakePort(object):

            def stopListening(oself):
                self._socket_paths.pop(socket_path)

        self._socket_paths[socket_path] = factory
        return FakePort()

    def connect_unix(self, path, factory):
        server = self._socket_paths.get(path)
        from landscape.lib.tests.test_amp import FakeConnector
        if server:
            connector = FakeConnector(factory, server)
            connector.connect()
        else:
            connector = object()  # Fake connector
            failure = Failure(ConnectError("No such file or directory"))
            factory.clientConnectionFailed(connector, failure)
        return connector

    def run(self):
        """Continuously advance this reactor until reactor.stop() is called."""
        self.fire("run")
        self._running = True
        while self._running:
            self.advance(self._calls[0][0])
        self.fire("stop")

    def stop(self):
        self._running = False

    def advance(self, seconds):
        """Advance this reactor C{seconds} into the future.

        This method is not part of the L{LandscapeReactor} API and is specific
        to L{FakeReactor}. It's meant to be used only in unit tests for
        advancing time and triggering the relevant scheduled calls (see
        also C{call_later} and C{call_every}).
        """
        while (self._calls and self._calls[0][0]
               <= self._current_time + seconds):
            call = self._calls.pop(0)
            # If we find a call within the time we're advancing,
            # before calling it, let's advance the time *just* to
            # when that call is expecting to be run, so that if it
            # schedules any calls itself they will be relative to
            # the correct time.
            seconds -= call[0] - self._current_time
            self._current_time = call[0]
            try:
                call[1](*call[2], **call[3])
            except Exception, e:
                logging.exception(e)
        self._current_time += seconds

    def _in_thread(self, callback, errback, f, args, kwargs):
        try:
            result = f(*args, **kwargs)
        except Exception, e:
            exc_info = sys.exc_info()
            if errback is None:
                self.call_in_main(logging.error, e, exc_info=exc_info)
            else:
                self.call_in_main(errback, *exc_info)
        else:
            if callback:
                self.call_in_main(callback, result)

    def _run_threaded_callbacks(self):
        while self._threaded_callbacks:
            try:
                self._threaded_callbacks.pop(0)()
            except Exception, e:
                logging.exception(e)

    def _hook_threaded_callbacks(self):
        id = self.call_every(0.5, self._run_threaded_callbacks)
        self._run_threaded_callbacks_id = id

    def _unhook_threaded_callbacks(self):
        self.cancel_call(self._run_threaded_callbacks_id)



from landscape.ui.tests.helpers import (
    ConfigurationProxyHelper, FakeGSettings, dbus_test_should_skip,
    dbus_skip_message)

if not dbus_test_should_skip:
    from landscape.ui.model.configuration.uisettings import UISettings
    import landscape.ui.model.configuration.state
    from landscape.ui.model.configuration.state import (
        ConfigurationModel, StateError, VirginState, InitialisedState,
        ModifiedState, MANAGEMENT_TYPE, HOSTED, LOCAL, HOSTED_LANDSCAPE_HOST,
        LANDSCAPE_HOST, COMPUTER_TITLE, ExitedState)
    from landscape.ui.model.configuration.mechanism import (
        PermissionDeniedByPolicy)
    from landscape.ui.constants import (
        CANONICAL_MANAGED, LOCAL_MANAGED, NOT_MANAGED)


from landscape.tests.helpers import LandscapeTest


class AuthenticationFailureTest(LandscapeTest):
    """
    Test that an authentication failure is handled correctly in the
    L{ConfigurationModel}.
    """
    helpers = [ConfigurationProxyHelper]

    def setUp(self):
        self.config_string = ""
        self.default_data = {"management-type": "canonical",
                             "computer-title": "",
                             "hosted-landscape-host": "",
                             "hosted-account-name": "",
                             "hosted-password": "",
                             "local-landscape-host": "",
                             "local-account-name": "",
                             "local-password": ""}
        landscape.ui.model.configuration.state.DEFAULT_DATA[COMPUTER_TITLE] \
            = "bound.to.lose"
        super(AuthenticationFailureTest, self).setUp()

    def test_failed_authentication(self):
        """
        Test that load returns False when authentication fails.
        """

        def fake_faily_load(arglist):
            """
            This simulates what you see if you click "Cancel" or give the wrong
            credentials 3 times when L{PolicyKit} challenges you.
            """
            raise PermissionDeniedByPolicy()

        def fake_exit_method():
            """
            Avoid raising a L{SystemExit} exception.
            """

        self.mechanism.load = fake_faily_load
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        self.assertFalse(model.load_data(asynchronous=False,
                                         exit_method=fake_exit_method))
        self.assertTrue(isinstance(model.get_state(), ExitedState))

    if dbus_test_should_skip:
        skip = dbus_skip_message


class ConfigurationModelTest(LandscapeTest):
    """
    Test the internal data handling of the L{ConfigurationModel} without
    loading external data.
    """

    helpers = [ConfigurationProxyHelper]

    def setUp(self):
        self.config_string = ""
        self.default_data = {"management-type": "canonical",
                             "computer-title": "",
                             "hosted-landscape-host": "",
                             "hosted-account-name": "",
                             "hosted-password": "",
                             "local-landscape-host": "",
                             "local-account-name": "",
                             "local-password": ""}
        landscape.ui.model.configuration.state.DEFAULT_DATA[COMPUTER_TITLE] \
            = "bound.to.lose"
        super(ConfigurationModelTest, self).setUp()

    def test_get(self):
        """
        Test that L{get} correctly extracts data from the internal data storage
        of the L{ConfigurationState}s associated with a L{ConfigurationModel}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        state = model.get_state()
        self.assertEqual(NOT_MANAGED, state.get(MANAGEMENT_TYPE))
        self.assertEqual(HOSTED_LANDSCAPE_HOST,
                         state.get(HOSTED, LANDSCAPE_HOST))
        self.assertRaises(TypeError, state.get, MANAGEMENT_TYPE, HOSTED,
                          LANDSCAPE_HOST)
        self.assertRaises(KeyError, state.get, LANDSCAPE_HOST)
        self.assertRaises(KeyError, state.get, MANAGEMENT_TYPE, LANDSCAPE_HOST)

    def test_set(self):
        """
        Test that L{set} correctly sets data in the internal data storage of
        the L{ConfigurationState}s associated with a L{ConfigurationModel}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        state = model.get_state()
        state.set(MANAGEMENT_TYPE, NOT_MANAGED)
        self.assertEqual(NOT_MANAGED, state.get(MANAGEMENT_TYPE))
        state.set(MANAGEMENT_TYPE, CANONICAL_MANAGED)
        self.assertEqual(CANONICAL_MANAGED, state.get(MANAGEMENT_TYPE))
        state.set(MANAGEMENT_TYPE, LOCAL_MANAGED)
        self.assertEqual(LOCAL_MANAGED, state.get(MANAGEMENT_TYPE))
        self.assertEqual("", state.get(LOCAL, LANDSCAPE_HOST))
        state.set(LOCAL, LANDSCAPE_HOST, "goodison.park")
        self.assertEqual("goodison.park", state.get(LOCAL, LANDSCAPE_HOST))

    def test_virginal(self):
        """
        Test that the L{ConfigurationModel} is created with default data.  This
        should be managed via L{VirginState} (hence the name), but this should
        not be exposed and is not explicitly tested here (see
        L{StateTransitionTest}).
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        self.assertEqual(NOT_MANAGED, model.management_type)
        self.assertEqual(HOSTED_LANDSCAPE_HOST, model.hosted_landscape_host)
        self.assertEqual("bound.to.lose", model.computer_title)
        self.assertEqual("", model.local_landscape_host)
        self.assertEqual("", model.hosted_account_name)
        self.assertEqual("standalone", model.local_account_name)
        self.assertEqual("", model.hosted_password)

    def test_is_hosted_property(self):
        """
        Test we can use the L{is_hosted} property to set and get that data on
        the current L{ConfigurationState}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        model.load_data()
        self.assertEqual(CANONICAL_MANAGED, model.management_type)
        model.management_type = LOCAL_MANAGED
        self.assertEqual(LOCAL_MANAGED, model.management_type)
        model.management_type = NOT_MANAGED
        self.assertEqual(NOT_MANAGED, model.management_type)

    def test_computer_title_property(self):
        """
        Test that we can use the L{computer_title} property to set and get that
        data on the current L{ConfigurationState}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        model.load_data()
        self.assertEqual("bound.to.lose", model.computer_title)
        model.computer_title = "bound.to.win"
        self.assertEqual("bound.to.win", model.computer_title)

    def test_hosted_landscape_host_property(self):
        """
        Test we can use the L{hosted_landscape_host} property to set and get
        that data on the current L{ConfigurationState}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        self.assertEqual(HOSTED_LANDSCAPE_HOST, model.hosted_landscape_host)
        self.assertRaises(AttributeError, setattr, model,
                          "hosted_landscape_host", "foo")

    def test_hosted_account_name_property(self):
        """
        Test we can use the L{hosted_account_name} property to set and get
        that data on the current L{ConfigurationState}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        self.assertEqual("", model.hosted_account_name)
        model.hosted_account_name = "foo"
        self.assertEqual("foo", model.hosted_account_name)

    def test_hosted_password_property(self):
        """
        Test we can use the L{hosted_password} property to set and get
        that data on the current L{ConfigurationState}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        self.assertEqual("", model.hosted_password)
        model.hosted_password = "foo"
        self.assertEqual("foo", model.hosted_password)

    def test_local_landscape_host_property(self):
        """
        Test we can use the L{local_landscape_host} property to set and get
        that data on the current L{ConfigurationState}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        self.assertEqual("", model.local_landscape_host)
        model.local_landscape_host = "foo"
        self.assertEqual("foo", model.local_landscape_host)

    def test_local_account_name_property(self):
        """
        Test we can use the L{local_account_name} property to set and get
        that data on the current L{ConfigurationState}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        self.assertEqual("standalone", model.local_account_name)
        model.local_account_name = "foo"
        self.assertEqual("foo", model.local_account_name)

    def test_local_password_property(self):
        """
        Test we can use the L{local_password} property to set and get
        that data on the current L{ConfigurationState}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        self.assertEqual("", model.local_password)
        model.local_password = "foo"
        self.assertEqual("foo", model.local_password)

    def test_exit(self):
        """
        Test that we can cause the mechanism to exit.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        self.assertRaises(SystemExit, model.exit, asynchronous=False)

    if dbus_test_should_skip:
        skip = dbus_skip_message


class ConfigurationModelHostedTest(LandscapeTest):
    """
    Test the L{ConfigurationModel} is correctly initialised when the live
    configuration is set for a hosted account.

    Note the multilayer data loading:

         1. Internal state is defaulted.
         2. UISettings data is loaded.
         3. Live configuration is loaded.
    """

    helpers = [ConfigurationProxyHelper]

    default_data = {"management-type": "canonical",
                    "computer-title": "bound.to.lose",
                    "hosted-landscape-host": "landscape.canonical.com",
                    "hosted-account-name": "Sparklehorse",
                    "hosted-password": "Vivadixiesubmarinetransmissionplot",
                    "local-landscape-host": "the.local.machine",
                    "local-account-name": "CrazyHorse",
                    "local-password": "RustNeverSleeps"}

    def setUp(self):
        self.config_string = "[client]\n" \
            "data_path = /var/lib/landscape/client/\n" \
            "http_proxy = http://proxy.localdomain:3192\n" \
            "tags = a_tag\n" \
            "url = https://landscape.canonical.com/message-system\n" \
            "account_name = foo\n" \
            "registration_key = boink\n" \
            "computer_title = baz\n" \
            "https_proxy = https://proxy.localdomain:6192\n" \
            "ping_url = http://landscape.canonical.com/ping\n"

        super(ConfigurationModelHostedTest, self).setUp()

    def test_initialised_hosted(self):
        """
        Test the L{ConfigurationModel} is correctly initialised from a proxy
        and defaults with hosted data.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        model.load_data()
        self.assertEqual(CANONICAL_MANAGED, model.management_type)
        self.assertEqual("landscape.canonical.com",
                         model.hosted_landscape_host)
        self.assertEqual("the.local.machine", model.local_landscape_host)
        self.assertEqual("foo", model.hosted_account_name)
        self.assertEqual("CrazyHorse", model.local_account_name)
        self.assertEqual("boink", model.hosted_password)

    if dbus_test_should_skip:
        skip = dbus_skip_message


class ConfigurationModelLocalTest(LandscapeTest):

    helpers = [ConfigurationProxyHelper]

    default_data = {"management-type": "LDS",
                    "computer-title": "bound.to.lose",
                    "hosted-landscape-host": "landscape.canonical.com",
                    "hosted-account-name": "Sparklehorse",
                    "hosted-password": "Vivadixiesubmarinetransmissionplot",
                    "local-landscape-host": "the.local.machine",
                    "local-account-name": "CrazyHorse",
                    "local-password": "RustNeverSleeps"}

    def setUp(self):
        self.config_string = "[client]\n" \
            "data_path = /var/lib/landscape/client/\n" \
            "http_proxy = http://proxy.localdomain:3192\n" \
            "tags = a_tag\n" \
            "url = https://landscape.localdomain/message-system\n" \
            "account_name = foo\n" \
            "registration_key = boink\n" \
            "computer_title = baz\n" \
            "https_proxy = \n" \
            "ping_url = http://landscape.localdomain/ping\n"

        super(ConfigurationModelLocalTest, self).setUp()

    def test_initialised_local(self):
        """
        Test the L{ConfigurationModel} is correctly initialised from a proxy
        and defaults with local data.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        model.load_data()
        self.assertEqual(LOCAL_MANAGED, model.management_type)
        self.assertEqual("landscape.canonical.com",
                         model.hosted_landscape_host)
        self.assertEqual("landscape.localdomain", model.local_landscape_host)
        self.assertEqual("Sparklehorse", model.hosted_account_name)
        self.assertEqual("foo", model.local_account_name)
        self.assertEqual("Vivadixiesubmarinetransmissionplot",
                         model.hosted_password)

    if dbus_test_should_skip:
        skip = dbus_skip_message


class StateTransitionTest(LandscapeTest):
    """
    Test that we make the correct state transitions when taking actions on the
    L{ConfigurationModel}.
    """

    helpers = [ConfigurationProxyHelper]

    def setUp(self):
        self.config_string = ""
        self.default_data = {
            "management-type": "canonical",
            "computer-title": "bound.to.lose",
            "hosted-landscape-host": "landscape.canonical.com",
            "hosted-account-name": "Sparklehorse",
            "hosted-password": "Vivadixiesubmarinetransmissionplot",
            "local-landscape-host": "the.local.machine",
            "local-account-name": "CrazyHorse",
            "local-password": "RustNeverSleeps"}
        super(StateTransitionTest, self).setUp()

    def test_load_data_transitions(self):
        """
        Test that the L{ConfigurationModel} correctly changes state as we call
        L{load_data}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        self.assertTrue(isinstance(model.get_state(), VirginState))
        model.load_data()
        self.assertTrue(isinstance(model.get_state(), InitialisedState))
        initialised = model.get_state()
        model.load_data()
        self.assertTrue(isinstance(model.get_state(), InitialisedState))
        self.assertIs(initialised, model.get_state())

    def test_modifying_a_virgin_raises(self):
        """
        Test that attempting a L{modify} a L{ConfigurationModel} in
        L{VirginState} raises a L{StateError}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        self.assertRaises(StateError, model.modify)

    def test_initialised_state_is_modifiable(self):
        """
        Test that the L{ConfigurationModel} transitions to L{ModifiedState}
        whenever L{modify} is called on it in L{InitialisedState}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        model.load_data()
        self.assertEqual(CANONICAL_MANAGED, model.management_type)
        model.management_type = LOCAL_MANAGED
        self.assertEqual(LOCAL_MANAGED, model.management_type)
        model.modify()
        self.assertTrue(isinstance(model.get_state(), ModifiedState))
        self.assertEqual(LOCAL_MANAGED, model.management_type)

    def test_modified_state_is_modifiable(self):
        """
        Test that the L{ConfigurationModel} transitions to L{ModifiedState}
        whenever L{modify} is called on it in L{ModifiedState}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        model.load_data()
        model.modify()
        self.assertTrue(isinstance(model.get_state(), ModifiedState))
        model.modify()
        self.assertTrue(isinstance(model.get_state(), ModifiedState))

    def test_reverting_a_virgin_raises(self):
        """
        Test that calling L{revert} on a L{ConfigurationModel} in
        L{VirginState} raises a L{StateError}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        self.assertRaises(StateError, model.revert)

    def test_initialiased_state_is_unrevertable(self):
        """
        Test that calling L{revert} on a L{ConfigurationModel} in
        L{InitialisedState} raises a L{StateError}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        model.load_data()
        self.assertRaises(StateError, model.revert)

    def test_modified_state_is_revertable(self):
        """
        Test that a L{ConfigurationModel} in L{ModifiedState} can be
        transitioned via L{revert} to L{InitialisedState}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        model.load_data()
        model.modify()
        model.revert()
        self.assertTrue(isinstance(model.get_state(), InitialisedState))

    def test_reverting_reverts_data(self):
        """
        Test that transitioning via L{revert} causes the original
        L{InitialisedState} to be restored.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        model.load_data()
        self.assertEqual(HOSTED_LANDSCAPE_HOST, model.hosted_landscape_host)
        self.assertEqual("CrazyHorse", model.local_account_name)
        model.local_account_name = "bar"
        model.modify()
        self.assertEqual("bar", model.local_account_name)
        model.revert()
        self.assertTrue(isinstance(model.get_state(), InitialisedState))
        self.assertEqual("CrazyHorse", model.local_account_name)

    def test_persisting_a_virgin_raises(self):
        """
        Test that a L{ConfigurationModel} in L{VirginState} will raise a
        L{StateError} when you attempt to transition it with L{persist}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        self.assertRaises(StateError, model.persist)

    def test_persisting_initialised_state_raises(self):
        """
        Test that a L{ConfigurationModel} in L{IntialisedState} will raise a
        L{StateError} when you attempt to transition it with L{persist}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        model.load_data()
        self.assertRaises(StateError, model.persist)

    def test_persisting_modified_is_allowed(self):
        """
        Test that a L{ConfigurationModel} in L{ModifiedState} will allow itself
        to be transitioned with L{persist}.
        """
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        model.load_data()
        model.modify()
        model.persist()
        self.assertTrue(isinstance(model.get_state(), InitialisedState))

    def test_persisting_saves_data_to_uisettings(self):
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        model.load_data()
        self.assertEqual(CANONICAL_MANAGED, uisettings.get_management_type())
        self.assertEqual("Sparklehorse", uisettings.get_hosted_account_name())
        self.assertEqual("Vivadixiesubmarinetransmissionplot",
                        uisettings.get_hosted_password())
        self.assertEqual("the.local.machine",
                         uisettings.get_local_landscape_host())
        self.assertEqual("CrazyHorse", uisettings.get_local_account_name())
        self.assertEqual("RustNeverSleeps", uisettings.get_local_password())
        model.management_type = LOCAL_MANAGED
        model.hosted_account_name = "ThomasPaine"
        model.hosted_password = "TheAgeOfReason"
        model.local_landscape_host = "another.local.machine"
        model.local_account_name = "ThomasHobbes"
        model.local_password = "TheLeviathan"
        model.modify()
        self.assertTrue(isinstance(model.get_state(), ModifiedState))
        model.persist()
        self.assertTrue(isinstance(model.get_state(), InitialisedState))
        self.assertEqual(LOCAL_MANAGED, uisettings.get_management_type())
        self.assertEqual("ThomasPaine", uisettings.get_hosted_account_name())
        self.assertEqual("TheAgeOfReason", uisettings.get_hosted_password())
        self.assertEqual("another.local.machine",
                         uisettings.get_local_landscape_host())
        self.assertEqual("ThomasHobbes", uisettings.get_local_account_name())
        self.assertEqual("TheLeviathan", uisettings.get_local_password())

    def test_any_transition_on_exited_state_raises(self):
        """
        Test that we cannot transition the L{ExitedState} at all.
        """

        def fake_exit():
            """
            This just avoids raising L{exceptions.SysExit} during __init__.
            """

        state = ExitedState(None, None, None, exit_method=fake_exit)
        self.assertRaises(StateError, state.load_data)
        self.assertRaises(StateError, state.modify)
        self.assertRaises(StateError, state.persist)
        self.assertRaises(StateError, state.revert)

    if dbus_test_should_skip:
        skip = dbus_skip_message


class StateTransitionWithExistingConfigTest(LandscapeTest):
    """
    Test that we handle existing configuration data correctly when
    transitioning through L{ConfigurationModel} states.
    """

    helpers = [ConfigurationProxyHelper]

    def setUp(self):
        self.config_string = (
            "[client]\n"
            "data_path = /var/lib/landscape/client/\n"
            "http_proxy = http://proxy.localdomain:3192\n"
            "tags = a_tag\n"
            "url = https://landscape.canonical.com/message-system\n"
            "account_name = Sparklehorse\n"
            "registration_key = Vivadixiesubmarinetransmissionplot\n"
            "computer_title = baz\n"
            "https_proxy = https://proxy.localdomain:6192\n"
            "ping_url = http://landscape.canonical.com/ping\n")
        self.default_data = {
            "management-type": "canonical",
            "computer-title": "bound.to.lose",
            "hosted-landscape-host": "landscape.canonical.com",
            "hosted-account-name": "Sparklehorse",
            "hosted-password": "Vivadixiesubmarinetransmissionplot",
            "local-landscape-host": "the.local.machine",
            "local-account-name": "CrazyHorse",
            "local-password": "RustNeverSleeps"}
        super(StateTransitionWithExistingConfigTest, self).setUp()

    def test_persisting_saves_data_to_proxy(self):
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        model.load_data()
        self.assertEqual("Sparklehorse", self.proxy.account_name)
        self.assertEqual("Vivadixiesubmarinetransmissionplot",
                        self.proxy.registration_key)
        model.management_type = LOCAL_MANAGED
        model.local_account_name = "ThomasPaine"
        model.local_password = "TheAgeOfReason"
        model.modify()
        self.assertTrue(isinstance(model.get_state(), ModifiedState))
        model.persist()
        self.assertTrue(isinstance(model.get_state(), InitialisedState))
        self.assertEqual(LOCAL_MANAGED, model.management_type)
        self.assertEqual("https://the.local.machine/message-system",
                         self.proxy.url)
        self.assertEqual("http://the.local.machine/ping", self.proxy.ping_url)
        self.assertEqual("ThomasPaine", self.proxy.account_name)
        self.assertEqual("TheAgeOfReason", self.proxy.registration_key)

    if dbus_test_should_skip:
        skip = dbus_skip_message



from twisted.internet.defer import succeed

from landscape.lib.sysstats import get_thermal_zones


class Temperature(object):

    def __init__(self, thermal_zone_path=None):
        self._thermal_zone_path = thermal_zone_path

    def register(self, sysinfo):
        self._sysinfo = sysinfo

    def run(self):
        temperature = None
        max_value = None
        for zone in get_thermal_zones(self._thermal_zone_path):
            if (zone.temperature_value is not None and
                (max_value is None or zone.temperature_value > max_value)):
                temperature = zone.temperature
                max_value = zone.temperature_value
        if temperature is not None:
            self._sysinfo.add_header("Temperature", temperature)
        return succeed(None)



import pprint
import os

from landscape.lib.persist import (
    path_string_to_tuple, path_tuple_to_string, Persist, RootedPersist,
    PickleBackend, PersistError, PersistReadOnlyError)
from landscape.tests.helpers import LandscapeTest


class PersistHelpersTest(LandscapeTest):

    paths = [
        ("ab", ("ab",)),
        ("ab.cd", ("ab", "cd")),
        ("ab.cd.de", ("ab", "cd", "de")),
        ("ab[0]", ("ab", 0)),
        ("ab[0][1]", ("ab", 0, 1)),
        ("ab.cd[1]", ("ab", "cd", 1)),
        ("ab[0].cd[1]", ("ab", 0, "cd", 1)),
        ("ab.cd.de[2]", ("ab", "cd", "de", 2)),
        ]

    def test_path_string_to_tuple(self):
        for path_string, path_tuple in self.paths:
            self.assertEqual(path_string_to_tuple(path_string), path_tuple)

    def test_path_string_to_tuple_error(self):
        self.assertRaises(PersistError, path_string_to_tuple, "ab[0][c]")

    def test_path_tuple_to_string(self):
        for path_string, path_tuple in self.paths:
            self.assertEqual(path_tuple_to_string(path_tuple), path_string)


class BasePersistTest(LandscapeTest):

    set_items = [
        ("ab", 1),
        ("ab", 2),
        ("cd.ef", 3.4),
        ("cd.gh", "4"),
        ("cd.gh", "5"),
        ("cd.ij.kl", (1, 2.3, "4", [5], (6,))),
        ("cd.ij.mn", [1, 2.3, "4", [5], (6,)]),
        ("cd.ij.op[1]", 0),
        ("cd.ij.op[1]", 1),
        ("cd.ij.op[2]", 2),
        ("qr", {"s": {"t": "u"}}),
        ("v", [0, {}, 2]),
        ("v[1].v", "woot"),
        ]

    set_result = {
                  "ab": 2,
                  "cd": {
                         "ef": 3.4,
                         "gh": "5",
                         "ij": {
                                "kl": (1, 2.3, "4", [5], (6,)),
                                "mn": [1, 2.3, "4", [5], (6,)],
                                "op": [0, 1, 2]
                               },
                         },
                   "qr": {"s": {"t": "u"}},
                   "v": [0, {"v": "woot"}, 2],
                 }

    get_items = [
                 ("ab", 2),
                 ("cd.ef", 3.4),
                 ("cd.gh", "5"),
                 ("cd.ij.kl", (1, 2.3, "4", [5], (6,))),
                 ("cd.ij.kl[3]", [5]),
                 ("cd.ij.kl[3][0]", 5),
                 ("cd.ij.mn", [1, 2.3, "4", [5], (6,)]),
                 ("cd.ij.mn.4", "4"),
                 ("cd.ij.mn.5", None),
                 ("cd.ij.op", [0, 1, 2]),
                 ("cd.ij.op[0]", 0),
                 ("cd.ij.op[1]", 1),
                 ("cd.ij.op[2]", 2),
                 ("cd.ij.op[3]", None),
                 ("qr", {"s": {"t": "u"}}),
                 ("qr.s", {"t": "u"}),
                 ("qr.s.t", "u"),
                 ("x", None),
                 ("x.y.z", None),
                ]

    add_items = [
                 ("ab", 1),
                 ("ab", 2.3),
                 ("ab", "4"),
                 ("ab", [5]),
                 ("ab", (6,)),
                 ("ab", {}),
                 ("ab[5].cd", "foo"),
                 ("ab[5].cd", "bar"),
                ]

    add_result = {
                  "ab": [1, 2.3, "4", [5], (6,), {"cd": ["foo", "bar"]}],
                 }

    def setUp(self):
        LandscapeTest.setUp(self)
        self.persist = self.build_persist()

    def tearDown(self):
        del self.persist
        LandscapeTest.tearDown(self)

    def build_persist(self, *args, **kwargs):
        return Persist(*args, **kwargs)

    def format(self, result, expected):
        repr_result = pprint.pformat(result)
        repr_expected = pprint.pformat(expected)
        return "\nResult:\n%s\nExpected:\n%s\n" % (repr_result, repr_expected)


class GeneralPersistTest(BasePersistTest):

    def test_set(self):
        for path, value in self.set_items:
            self.persist.set(path, value)
        result = self.persist.get((), hard=True)
        self.assertEqual(result, self.set_result,
                         self.format(result, self.set_result))

    def test_set_tuple_paths(self):
        for path, value in self.set_items:
            self.persist.set(path_string_to_tuple(path), value)
        result = self.persist.get((), hard=True)
        self.assertEqual(result, self.set_result,
                         self.format(result, self.set_result))

    def test_set_from_result(self):
        for path in self.set_result:
            self.persist.set(path, self.set_result[path])
        result = self.persist.get((), hard=True)
        self.assertEqual(result, self.set_result,
                         self.format(result, self.set_result))

    def test_get(self):
        for path in self.set_result:
            self.persist.set(path, self.set_result[path])
        for path, value in self.get_items:
            self.assertEqual(self.persist.get(path), value)

    def test_get_tuple_paths(self):
        for path in self.set_result:
            self.persist.set(path_string_to_tuple(path), self.set_result[path])
        for path, value in self.get_items:
            self.assertEqual(self.persist.get(path), value)

    def test_add(self):
        for path, value in self.add_items:
            self.persist.add(path, value)
        result = self.persist.get((), hard=True)
        self.assertEqual(result, self.add_result,
                         self.format(result, self.add_result))

    def test_add_unique(self):
        self.persist.add("a", "b")
        self.assertEqual(self.persist.get("a"), ["b"])
        self.persist.add("a", "b")
        self.assertEqual(self.persist.get("a"), ["b", "b"])
        self.persist.add("a", "b", unique=True)
        self.assertEqual(self.persist.get("a"), ["b", "b"])
        self.persist.add("a", "c", unique=True)
        self.assertEqual(self.persist.get("a"), ["b", "b", "c"])

    def test_keys(self):
        self.persist.set("a", {"b": 1, "c": {"d": 2}, "e": list("foo")})
        keys = self.persist.keys
        self.assertEqual(set(keys((), hard=True)), set(["a"]))
        self.assertEqual(set(keys("a")), set(["b", "c", "e"]))
        self.assertEqual(set(keys("a.d")), set([]))
        self.assertEqual(set(keys("a.e")), set([0, 1, 2]))
        self.assertEqual(set(keys("a.f")), set([]))
        self.assertRaises(PersistError, keys, "a.b")

    def test_has(self):
        self.persist.set("a", {"b": 1, "c": {"d": 2}, "e": list("foo")})
        has = self.persist.has
        self.assertTrue(has("a"))
        self.assertTrue(has(("a", "b")))
        self.assertTrue(has("a.c"))
        self.assertTrue(has("a.c", "d"))
        self.assertTrue(has("a.c.d"))
        self.assertTrue(has("a.e"))
        self.assertTrue(has("a.e[0]"))
        self.assertTrue(has("a.e", "f"))
        self.assertTrue(has("a.e", "o"))
        self.assertFalse(has("b"))
        self.assertFalse(has("a.f"))
        self.assertFalse(has("a.c.f"))
        self.assertFalse(has("a.e[3]"))
        self.assertFalse(has("a.e", "g"))
        self.assertRaises(PersistError, has, "a.b.c")

    def test_remove(self):
        self.persist.set("a", {"b": [1], "c": {"d": 2}, "e": list("foot")})
        get = self.persist.get
        has = self.persist.has
        remove = self.persist.remove

        self.assertFalse(remove("a.f"))

        self.assertRaises(PersistError, remove, "a.c.d.e")

        self.assertTrue(remove(("a", "e", "o")))
        self.assertEqual(get("a.e"), ["f", "t"])

        self.assertFalse(remove("a.e[2]"))
        self.assertEqual(get("a.e"), ["f", "t"])

        self.assertTrue(remove("a.e[1]"))
        self.assertEqual(get("a.e"), ["f"])

        self.assertTrue(remove("a.e", "f"))
        self.assertFalse(has("a.e"))

        self.assertFalse(remove("a.b[1]"))
        self.assertEqual(get("a.b"), [1])

        self.assertTrue(remove("a.b", 1))
        self.assertFalse(has("a.b"))

        self.assertTrue(remove("a.c"))
        self.assertFalse(has("a.c"))

        self.assertFalse(has("a"))

    def test_move(self):
        self.persist.set("a", {"b": [1], "c": {"d": 2}})

        move = self.persist.move
        get = self.persist.get

        self.assertTrue(move("a.b", "a.c.b"))
        self.assertEqual(get("a"), {"c": {"b": [1], "d": 2}})

        self.assertTrue(move("a.c.b[0]", "a.c.b"))
        self.assertEqual(get("a"), {"c": {"b": 1, "d": 2}})

        self.assertTrue(move(("a", "c", "b"), ("a", "c", "b", 0)))
        self.assertEqual(get("a"), {"c": {"b": [1], "d": 2}})

    def test_copy_values_on_set(self):
        d = {"b": 1}
        d_orig = d.copy()
        self.persist.set("a", d)
        d["c"] = 2
        self.assertEqual(self.persist.get("a"), d_orig)

    def test_copy_values_on_add(self):
        d = {"b": 1}
        d_orig = d.copy()
        self.persist.add("a", d)
        d["c"] = 2
        self.assertEqual(self.persist.get("a[0]"), d_orig)

    def test_copy_values_on_get(self):
        self.persist.set("a", {"b": 1})
        d = self.persist.get("a")
        d_orig = d.copy()
        d["c"] = 2
        self.assertEqual(self.persist.get("a"), d_orig)

    def test_root_at(self):
        rooted = self.persist.root_at("my-module")
        rooted.set("option", 1)
        self.assertEqual(self.persist.get("my-module.option"), 1)


class SaveLoadPersistTest(BasePersistTest):

    def test_readonly(self):
        self.assertFalse(self.persist.readonly)
        self.persist.readonly = True
        self.assertTrue(self.persist.readonly)
        self.persist.readonly = False
        self.assertFalse(self.persist.readonly)

        self.persist.readonly = True
        self.assertRaises(PersistReadOnlyError, self.persist.set, "ab", 2)
        self.assertRaises(PersistReadOnlyError, self.persist.add, "ab", 3)
        self.assertRaises(PersistReadOnlyError, self.persist.remove, "ab", 4)
        self.assertRaises(PersistReadOnlyError, self.persist.move, "ab", "cd")

        for keyword in ["weak", "soft"]:
            kwargs = {keyword: True}
            self.persist.set("ab", 2, **kwargs)
            self.persist.add("cd", 2, **kwargs)
            self.persist.remove("ab", **kwargs)
            self.persist.move("cd", "ef", **kwargs)

    def test_assert_writable(self):
        self.persist.assert_writable()
        self.persist.set("ab", 1)
        self.persist.readonly = True
        self.assertRaises(PersistReadOnlyError, self.persist.assert_writable)

    def test_modified(self):
        self.assertFalse(self.persist.modified)
        self.persist.set("ab", 1)
        self.assertTrue(self.persist.modified)
        self.persist.reset_modified()
        self.assertFalse(self.persist.modified)
        self.persist.add("cd", 2)
        self.assertTrue(self.persist.modified)
        self.persist.reset_modified()
        self.assertFalse(self.persist.modified)
        self.persist.remove("ab")
        self.assertTrue(self.persist.modified)
        self.persist.reset_modified()
        self.assertFalse(self.persist.modified)
        self.persist.move("cd", "ef")
        self.assertTrue(self.persist.modified)

    def test_save_and_load(self):
        for path in self.set_result:
            self.persist.set(path, self.set_result[path])

        filename = self.makeFile()
        self.persist.save(filename)

        persist = self.build_persist()
        persist.load(filename)

        result = persist.get((), hard=True)
        self.assertEqual(result, self.set_result,
                         self.format(result, self.set_result))

    def test_save_on_unexistent_dir(self):
        dirname = self.makeFile()
        filename = os.path.join(dirname, "foobar")

        self.assertFalse(os.path.exists(dirname))
        self.persist.save(filename)
        self.assertTrue(os.path.isfile(filename))

    def test_save_creates_backup(self):
        filename = self.makePersistFile("foobar")
        filename_old = filename + ".old"

        self.assertFalse(os.path.exists(filename_old))
        self.persist.save(filename)
        self.assertTrue(os.path.exists(filename_old))

    def test_save_to_default_file(self):
        """
        Persist can be constructed with a filename, and Persist.save with no
        arguments will write to that filename.
        """
        filename = self.makeFile()
        persist = self.build_persist(filename=filename)
        self.assertFalse(os.path.exists(filename))
        persist.save()
        self.assertTrue(os.path.exists(filename))

    def test_save_to_no_default_file(self):
        """
        If no default filename was given, calling Persist.save with no
        arguments will raise a PersistError.
        """
        self.assertRaises(PersistError, self.persist.save)

    def test_load_default_file(self):
        """
        If a Persist is created with a default filename, and the filename
        exists, it will be loaded.
        """
        filename = self.makeFile()
        persist = self.build_persist(filename=filename)
        persist.set("foo", "bar")
        persist.save()

        persist = self.build_persist(filename=filename)
        self.assertEqual(persist.get("foo"), "bar")

    def test_load_restores_backup(self):
        filename = self.makePersistFile("foobar")
        filename_old = filename + ".old"

        self.persist.set("a", 1)
        self.persist.save(filename_old)

        persist = self.build_persist()
        persist.load(filename)

        self.assertEqual(persist.get("a"), 1)

    def test_load_empty_files_wont_break(self):
        filename = self.makeFile("")
        self.persist.load(filename)

    def test_load_empty_files_restore_backup(self):
        """
        If the current file is empty, it tries to load the old one if it
        exists.
        """
        filename = self.makeFile("")
        filename_old = filename + ".old"

        self.persist.set("a", 1)
        self.persist.save(filename_old)

        persist = self.build_persist()
        persist.load(filename)

        self.assertEqual(persist.get("a"), 1)

    def test_non_existing_raise_error(self):
        """
        Trying to load a file that doesn't exist result in a L{PersistError}.
        """
        persist = self.build_persist()
        self.assertRaises(PersistError, persist.load, "/nonexistent")

    def test_non_existing_restore_backup(self):
        """
        If the file doesn't exist, it tries to load the old one if present and
        valid.
        """
        filename = self.makeFile("")
        filename_old = filename + ".old"
        os.unlink(filename)

        self.persist.set("a", 1)
        self.persist.save(filename_old)

        persist = self.build_persist()
        persist.load(filename)

        self.assertEqual(persist.get("a"), 1)


class PicklePersistTest(GeneralPersistTest, SaveLoadPersistTest):

    def build_persist(self, *args, **kwargs):
        return Persist(PickleBackend(), *args, **kwargs)


class RootedPersistTest(GeneralPersistTest):

    def build_persist(self, *args, **kwargs):
        return RootedPersist(Persist(), "root.path", *args, **kwargs)

    def test_readonly(self):
        self.assertFalse(self.persist.readonly)
        self.assertRaises(AttributeError,
                          setattr, self.persist, "readonly", True)
        self.persist.parent.readonly = True
        self.assertTrue(self.persist.readonly)

    def test_assert_writable(self):
        self.persist.assert_writable()
        self.persist.set("ab", 1)
        self.persist.parent.readonly = True
        self.assertRaises(PersistReadOnlyError, self.persist.assert_writable)

    def test_modified(self):
        self.assertFalse(self.persist.modified)
        self.persist.set("ab", 1)
        self.assertTrue(self.persist.modified)
        self.persist.parent.reset_modified()
        self.assertFalse(self.persist.modified)
        self.persist.add("cd", 2)
        self.assertTrue(self.persist.modified)
        self.persist.parent.reset_modified()
        self.assertFalse(self.persist.modified)
        self.persist.remove("ab")
        self.assertTrue(self.persist.modified)
        self.persist.parent.reset_modified()
        self.assertFalse(self.persist.modified)
        self.persist.move("cd", "ef")
        self.assertTrue(self.persist.modified)



import os
import unittest
import signal
import tarfile

from twisted.internet import reactor
from twisted.internet.defer import succeed, fail, Deferred

from landscape.lib.gpg import InvalidGPGSignature
from landscape.lib.fetch import HTTPCodeError
from landscape.package.store import PackageStore
from landscape.package.releaseupgrader import (
    ReleaseUpgrader, ReleaseUpgraderConfiguration, main)
from landscape.tests.helpers import (
    LandscapeTest, BrokerServiceHelper, LogKeeperHelper, EnvironSaverHelper)
from landscape.manager.manager import SUCCEEDED, FAILED


class ReleaseUpgraderConfigurationTest(unittest.TestCase):

    def test_upgrade_tool_directory(self):
        """
        L{ReleaseUpgraderConfiguration.upgrade_tool_directory} returns the
        path to the directory holding the fetched upgrade-tool files.
        """
        config = ReleaseUpgraderConfiguration()
        self.assertEqual(config.upgrade_tool_directory,
                         os.path.join(config.package_directory,
                                      "upgrade-tool"))


class ReleaseUpgraderTest(LandscapeTest):

    helpers = [LogKeeperHelper,
               EnvironSaverHelper, BrokerServiceHelper]

    def setUp(self):
        super(ReleaseUpgraderTest, self).setUp()
        self.config = ReleaseUpgraderConfiguration()
        self.config.data_path = self.makeDir()
        os.mkdir(self.config.package_directory)
        os.mkdir(self.config.upgrade_tool_directory)
        self.store = PackageStore(self.makeFile())
        self.upgrader = ReleaseUpgrader(self.store, None,
                                        self.remote, self.config)
        service = self.broker_service
        service.message_store.set_accepted_types(["operation-result"])

    def get_pending_messages(self):
        return self.broker_service.message_store.get_pending_messages()

    def test_fetch(self):
        """
        L{ReleaseUpgrader.fetch} fetches the upgrade tool tarball and signature
        from the given URLs.
        """
        tarball_url = "http://some/where/karmic.tar.gz"
        signature_url = "http://some/where/karmic.tar.gz.gpg"

        os.rmdir(self.config.upgrade_tool_directory)
        fetch_mock = self.mocker.replace("landscape.lib.fetch.fetch_async")
        fetch_mock(tarball_url)
        self.mocker.result(succeed("tarball"))
        fetch_mock(signature_url)
        self.mocker.result(succeed("signature"))

        self.mocker.replay()
        result = self.upgrader.fetch(tarball_url,
                                     signature_url)

        def check_result(ignored):
            directory = self.config.upgrade_tool_directory
            self.assertFileContent(
                os.path.join(directory, "karmic.tar.gz"), "tarball")
            self.assertFileContent(
                os.path.join(directory, "karmic.tar.gz.gpg"), "signature")
            self.assertIn("INFO: Successfully fetched upgrade-tool files",
                          self.logfile.getvalue())

        result.addCallback(check_result)
        return result

    def test_fetch_with_errors(self):
        """
        L{ReleaseUpgrader.fetch} logs a warning in case any of the upgrade tool
        files fails to be fetched.
        """
        tarball_url = "http://some/where/karmic.tar.gz"
        signature_url = "http://some/where/karmic.tar.gz.gpg"

        fetch_mock = self.mocker.replace("landscape.lib.fetch.fetch_async")
        fetch_mock(tarball_url)
        self.mocker.result(succeed("tarball"))
        fetch_mock(signature_url)
        self.mocker.result(fail(HTTPCodeError(404, "not found")))
        self.mocker.replay()

        result = self.upgrader.fetch(tarball_url,
                                     signature_url)

        def check_failure(failure):
            self.assertIn("WARNING: Couldn't fetch file from %s (Server return"
                          "ed HTTP code 404)" % signature_url,
                          self.logfile.getvalue())
            self.assertIn("WARNING: Couldn't fetch all upgrade-tool files",
                          self.logfile.getvalue())

        result.addCallback(self.fail)
        result.addErrback(check_failure)
        return result

    def test_verify(self):
        """
        L{ReleaseUpgrader.verify} verifies the upgrade tool tarball against
        its signature.
        """
        tarball_filename = "/some/tarball"
        signature_filename = "/some/signature"

        gpg_verify_mock = self.mocker.replace("landscape.lib.gpg.gpg_verify")
        gpg_verify_mock(tarball_filename, signature_filename)
        self.mocker.result(succeed(True))
        self.mocker.replay()

        result = self.upgrader.verify(tarball_filename, signature_filename)

        def check_result(ignored):
            self.assertIn("INFO: Successfully verified upgrade-tool tarball",
                          self.logfile.getvalue())

        result.addCallback(check_result)
        return result

    def test_verify_invalid_signature(self):
        """
        L{ReleaseUpgrader.verify} logs a warning in case the tarball signature
        is not valid.
        """
        tarball_filename = "/some/tarball"
        signature_filename = "/some/signature"

        gpg_verify_mock = self.mocker.replace("landscape.lib.gpg.gpg_verify")
        gpg_verify_mock(tarball_filename, signature_filename)
        self.mocker.result(fail(InvalidGPGSignature("gpg error")))
        self.mocker.replay()

        result = self.upgrader.verify(tarball_filename, signature_filename)

        def check_failure(failure):
            self.assertIn("WARNING: Invalid signature for upgrade-tool "
                          "tarball: gpg error", self.logfile.getvalue())

        result.addCallback(self.fail)
        result.addErrback(check_failure)
        return result

    def test_extract(self):
        """
        The L{ReleaseUpgrader.extract} method extracts the upgrade-tool tarball
        in the proper directory.
        """
        original_filename = self.makeFile("data\n")
        tarball_filename = self.makeFile()
        tarball = tarfile.open(tarball_filename, "w:gz")
        tarball.add(original_filename, arcname="file")
        tarball.close()

        result = self.upgrader.extract(tarball_filename)

        def check_result(ignored):
            filename = os.path.join(self.config.upgrade_tool_directory, "file")
            self.assertTrue(os.path.exists(filename))
            self.assertFileContent(filename, "data\n")

        result.addCallback(check_result)
        return result

    def test_tweak_includes_landscape_ppa_in_mirrors(self):
        """
        The L{ReleaseUpgrader.tweak} method adds the Landscape PPA repository
        to the list of available mirrors.
        """
        mirrors_filename = os.path.join(self.config.upgrade_tool_directory,
                                        "mirrors.cfg")
        self.makeFile(path=mirrors_filename,
                      content="ftp://ftp.lug.ro/ubuntu/\n")

        def check_result(ignored):
            self.assertFileContent(mirrors_filename,
                                   "ftp://ftp.lug.ro/ubuntu/\n"
                                   "http://ppa.launchpad.net/landscape/"
                                   "trunk/ubuntu/\n")

        result = self.upgrader.tweak("hardy")
        result.addCallback(check_result)
        return result

    def test_default_logs_directory(self):
        """
        The default directory for the upgrade-tool logs is the system one.
        """
        self.assertEqual(self.upgrader.logs_directory,
                         "/var/log/dist-upgrade")

    def test_default_logs_limit(self):
        """
        The default read limit for the upgrade-tool logs is 100000 bytes.
        """
        self.assertEqual(self.upgrader.logs_limit, 100000)

    def test_make_operation_result_text(self):
        """
        L{ReleaseUpgrade.make_operation_result_text} aggregates the contents of
        the process standard output, error and log files.
        """
        self.upgrader.logs_directory = self.makeDir()
        self.makeFile(basename="main.log",
                      dirname=self.upgrader.logs_directory,
                      content="main log")
        self.makeFile(basename="apt.log",
                      dirname=self.upgrader.logs_directory,
                      content="apt log")
        text = self.upgrader.make_operation_result_text("stdout", "stderr")
        self.assertEqual(text,
                         "=== Standard output ===\n\n"
                         "stdout\n\n"
                         "=== Standard error ===\n\n"
                         "stderr\n\n"
                         "=== apt.log ===\n\n"
                         "apt log\n\n"
                         "=== main.log ===\n\n"
                         "main log\n\n")

    def test_make_operation_result_text_with_no_stderr(self):
        """
        L{ReleaseUpgrade.make_operation_result_text} skips the standard error
        if it's empty.
        """
        self.upgrader.logs_directory = self.makeDir()
        text = self.upgrader.make_operation_result_text("stdout", "")
        self.assertEqual(text,
                         "=== Standard output ===\n\n"
                         "stdout\n\n")

    def test_make_operation_result_text_only_considers_log_files(self):
        """
        L{ReleaseUpgrade.make_operation_result_text} only considers log files
        from the last upgrade-tool run, directories containing log files from
        an older run are skipped.
        """
        self.upgrader.logs_directory = self.makeDir()
        self.makeDir(dirname=self.upgrader.logs_directory)
        text = self.upgrader.make_operation_result_text("stdout", "stderr")
        self.assertEqual(text,
                         "=== Standard output ===\n\n"
                         "stdout\n\n"
                         "=== Standard error ===\n\n"
                         "stderr\n\n")

    def test_make_operation_result_text_trims_long_files(self):
        """
        L{ReleaseUpgrade.make_operation_result_text} only reads the last
        L{logs_limit} lines of a log file.
        """
        self.upgrader.logs_directory = self.makeDir()
        self.upgrader.logs_limit = 8
        self.makeFile(basename="main.log",
                      dirname=self.upgrader.logs_directory,
                      content="very long log")
        text = self.upgrader.make_operation_result_text("stdout", "stderr")
        self.assertEqual(text,
                         "=== Standard output ===\n\n"
                         "stdout\n\n"
                         "=== Standard error ===\n\n"
                         "stderr\n\n"
                         "=== main.log ===\n\n"
                         "long log\n\n")

    def test_upgrade(self):
        """
        The L{ReleaseUpgrader.upgrade} method spawns the appropropriate
        upgrade-tool script and reports the result.
        """
        self.upgrader.logs_directory = self.makeDir()
        upgrade_tool_directory = self.config.upgrade_tool_directory
        upgrade_tool_filename = os.path.join(upgrade_tool_directory, "karmic")
        fd = open(upgrade_tool_filename, "w")
        fd.write("#!/bin/sh\n"
                 "echo $@\n"
                 "echo FOO=$FOO\n"
                 "echo PWD=$PWD\n"
                 "echo out\n")
        fd.close()
        os.chmod(upgrade_tool_filename, 0755)
        env_backup = os.environ.copy()
        os.environ.clear()
        os.environ.update({"FOO": "bar"})
        deferred = Deferred()

        def do_test():

            result = self.upgrader.upgrade("karmic", 100)

            def check_result(ignored):
                self.assertIn("INFO: Queuing message with release upgrade "
                              "results to exchange urgently.",
                              self.logfile.getvalue())
                result_text = (u"=== Standard output ===\n\n"
                               "--frontend DistUpgradeViewNonInteractive\n"
                               "FOO=bar\n"
                               "PWD=%s\nout\n\n\n" % upgrade_tool_directory)
                self.assertMessages(self.get_pending_messages(),
                                    [{"type": "operation-result",
                                      "operation-id": 100,
                                      "status": SUCCEEDED,
                                      "result-text": result_text,
                                      "result-code": 0}])

            result.addCallback(check_result)
            result.chainDeferred(deferred)

        reactor.callWhenRunning(do_test)

        def cleanup(ignored):
            os.environ = env_backup
            return ignored

        return deferred.addBoth(cleanup)

    def test_upgrade_with_env_variables(self):
        """
        The L{ReleaseUpgrader.upgrade} method optionally sets environment
        variables to be passed to the upgrade-tool process.
        """
        self.upgrader.logs_directory = self.makeDir()
        upgrade_tool_directory = self.config.upgrade_tool_directory
        upgrade_tool_filename = os.path.join(upgrade_tool_directory, "karmic")
        fd = open(upgrade_tool_filename, "w")
        fd.write("#!/bin/sh\n"
                 "echo DEBUG_UPDATE_MANAGER=$DEBUG_UPDATE_MANAGER\n"
                 "echo RELEASE_UPRADER_ALLOW_THIRD_PARTY="
                 "$RELEASE_UPRADER_ALLOW_THIRD_PARTY\n")
        fd.close()
        os.chmod(upgrade_tool_filename, 0755)
        env_backup = os.environ.copy()
        os.environ.clear()
        deferred = Deferred()

        def do_test():

            result = self.upgrader.upgrade("karmic", 100,
                                           allow_third_party=True,
                                           debug=True)

            def check_result(ignored):
                result_text = (u"=== Standard output ===\n\n"
                               "DEBUG_UPDATE_MANAGER=True\n"
                               "RELEASE_UPRADER_ALLOW_THIRD_PARTY=True\n\n\n")
                self.assertMessages(self.get_pending_messages(),
                                    [{"type": "operation-result",
                                      "operation-id": 100,
                                      "status": SUCCEEDED,
                                      "result-text": result_text,
                                      "result-code": 0}])

            result.addCallback(check_result)
            result.chainDeferred(deferred)

        reactor.callWhenRunning(do_test)

        def cleanup(ignored):
            os.environ = env_backup
            return ignored

        return deferred.addBoth(cleanup)

    def test_upgrade_with_failure(self):
        """
        The L{ReleaseUpgrader.upgrade} sends a message with failed status
        field if the upgrade-tool exits with non-zero code.
        """
        self.upgrader.logs_directory = self.makeDir()
        upgrade_tool_directory = self.config.upgrade_tool_directory
        upgrade_tool_filename = os.path.join(upgrade_tool_directory, "karmic")
        fd = open(upgrade_tool_filename, "w")
        fd.write("#!/bin/sh\n"
                 "echo out\n"
                 "echo err >&2\n"
                 "exit 3")
        fd.close()
        os.chmod(upgrade_tool_filename, 0755)

        deferred = Deferred()

        def do_test():

            result = self.upgrader.upgrade("karmic", 100)

            def check_result(ignored):
                result_text = (u"=== Standard output ===\n\nout\n\n\n"
                               "=== Standard error ===\n\nerr\n\n\n")
                self.assertMessages(self.get_pending_messages(),
                                    [{"type": "operation-result",
                                      "operation-id": 100,
                                      "status": FAILED,
                                      "result-text": result_text,
                                      "result-code": 3}])

            result.addCallback(check_result)
            result.chainDeferred(deferred)

        reactor.callWhenRunning(do_test)
        return deferred

    def test_upgrade_with_open_child_fds(self):
        """
        The deferred returned by the L{ReleaseUpgrader.upgrade} method
        callbacks correctly even if the spawned upgrade-tool process forks
        and passes its files descriptors over to child processes we don't know
        about.
        """
        self.upgrader.logs_directory = self.makeDir()
        upgrade_tool_directory = self.config.upgrade_tool_directory
        upgrade_tool_filename = os.path.join(upgrade_tool_directory, "karmic")
        child_pid_filename = self.makeFile()
        fd = open(upgrade_tool_filename, "w")
        fd.write("#!/usr/bin/env python\n"
                 "import os\n"
                 "import time\n"
                 "import sys\n"
                 "if __name__ == '__main__':\n"
                 "    print 'First parent'\n"
                 "    pid = os.fork()\n"
                 "    if pid > 0:\n"
                 "        time.sleep(0.5)\n"
                 "        sys.exit(0)\n"
                 "    pid = os.fork()\n"
                 "    if pid > 0:\n"
                 "        fd = open('%s', 'w')\n"
                 "        fd.write(str(pid))\n"
                 "        fd.close()\n"
                 "        sys.exit(0)\n"
                 "    while True:\n"
                 "        time.sleep(2)\n" % child_pid_filename)
        fd.close()
        os.chmod(upgrade_tool_filename, 0755)
        os.environ.clear()
        os.environ.update({"FOO": "bar"})
        deferred = Deferred()

        def do_test():

            result = self.upgrader.upgrade("karmic", 100)

            def kill_child(how):
                fd = open(child_pid_filename, "r")
                child_pid = int(fd.read())
                fd.close()
                os.remove(child_pid_filename)
                try:
                    os.kill(child_pid, signal.SIGKILL)
                    self.assertEqual(how, "cleanly")
                    return child_pid
                except OSError:
                    pass

            force_kill_child = reactor.callLater(2, kill_child, "brutally")

            def check_result(ignored):
                force_kill_child.cancel()
                self.assertIn("INFO: Queuing message with release upgrade "
                              "results to exchange urgently.",
                              self.logfile.getvalue())
                kill_child("cleanly")
                result_text = self.get_pending_messages()[0]["result-text"]
                self.assertIn("First parent\n", result_text)

            result.addCallback(check_result)
            result.chainDeferred(deferred)

        reactor.callWhenRunning(do_test)

        def cleanup(ignored):
            self.assertFalse(os.path.exists(child_pid_filename))
            return ignored

        return deferred.addBoth(cleanup)

    def test_finish(self):
        """
        The L{ReleaseUpgrader.finish} method wipes the upgrade-tool directory
        and spawn the package-reporter, to inform the server of the changed
        packages.
        """
        upgrade_tool_directory = self.config.upgrade_tool_directory
        open(os.path.join(upgrade_tool_directory, "somefile"), "w").close()
        os.mkdir(os.path.join(upgrade_tool_directory, "somedir"))

        reporter_filename = self.makeFile("#!/bin/sh\n"
                                          "echo $@\n"
                                          "echo $(pwd)\n")
        os.chmod(reporter_filename, 0755)

        find_reporter_mock = self.mocker.replace("landscape.package.reporter."
                                                 "find_reporter_command")
        find_reporter_mock()
        self.mocker.result(reporter_filename)
        self.mocker.replay()

        deferred = Deferred()

        def do_test():

            result = self.upgrader.finish()

            def check_result((out, err, code)):
                self.assertFalse(os.path.exists(upgrade_tool_directory))
                self.assertEqual(out, "--force-apt-update\n%s\n"
                                  % os.getcwd())
                self.assertEqual(err, "")
                self.assertEqual(code, 0)

            result.addCallback(check_result)
            result.chainDeferred(deferred)

        reactor.callWhenRunning(do_test)
        return deferred

    def test_finish_as_root(self):
        """
        If the release-upgrader process is run as root, as it alwyas should,
        the L{ReleaseUpgrader.finish} method spawns the package-reporter with
        the landscape uid and gid.
        """

        find_reporter_mock = self.mocker.replace("landscape.package.reporter."
                                                 "find_reporter_command")
        find_reporter_mock()
        self.mocker.result("reporter")

        getuid_mock = self.mocker.replace("os.getuid")
        getuid_mock()
        self.mocker.result(0)

        getpwnam_mock = self.mocker.replace("pwd.getpwnam")
        getpwnam_mock("landscape")

        class FakePwNam(object):
            pw_uid = 1234

        self.mocker.result(FakePwNam())

        getgrnam_mock = self.mocker.replace("grp.getgrnam")
        getgrnam_mock("landscape")

        class FakeGrNam(object):
            gr_gid = 5678

        self.mocker.result(FakeGrNam())

        spawn_process_calls = []

        def spawn_process(pp, reporter, args=None, uid=None, gid=None,
                          path=None, env=None, usePTY=None):
            self.assertEqual(uid, 1234)
            self.assertEqual(gid, 5678)
            spawn_process_calls.append(True)

        saved_spawn_process = reactor.spawnProcess
        reactor.spawnProcess = spawn_process

        self.mocker.replay()

        try:
            self.upgrader.finish()
        finally:
            reactor.spawnProcess = saved_spawn_process
        self.assertEqual(spawn_process_calls, [True])

    def test_finish_with_config_file(self):
        """
        The L{ReleaseUpgrader.finish} method passes over to the reporter the
        configuration file the release-upgrader was called with.
        """
        reporter_filename = self.makeFile("#!/bin/sh\necho $@\n")
        os.chmod(reporter_filename, 0755)
        self.config.config = "/some/config"
        find_reporter_mock = self.mocker.replace("landscape.package.reporter."
                                                 "find_reporter_command")
        find_reporter_mock()
        self.mocker.result(reporter_filename)
        self.mocker.replay()

        deferred = Deferred()

        def do_test():

            result = self.upgrader.finish()

            def check_result((out, err, code)):
                self.assertEqual(out, "--force-apt-update "
                                       "--config=/some/config\n")
                self.assertEqual(err, "")
                self.assertEqual(code, 0)

            result.addCallback(check_result)
            result.chainDeferred(deferred)

        reactor.callWhenRunning(do_test)
        return deferred

    def test_handle_release_upgrade(self):
        """
        The L{ReleaseUpgrader.handle_release_upgrade} method calls the other
        helper methods in the right order and with the right arguments.
        """
        calls = []
        upgrade_tool_directory = self.config.upgrade_tool_directory

        def fetch(tarball_url, signature_url):
            self.assertEqual(tarball_url, "http://some/tarball")
            self.assertEqual(signature_url, "http://some/sign")
            calls.append("fetch")
            return succeed(None)

        def verify(tarball_filename, signature_filename):
            self.assertEqual(tarball_filename,
                             os.path.join(upgrade_tool_directory, "tarball"))
            self.assertEqual(signature_filename,
                             os.path.join(upgrade_tool_directory, "sign"))
            calls.append("verify")

        def extract(filename_tarball):
            self.assertEqual(filename_tarball,
                             os.path.join(upgrade_tool_directory, "tarball"))
            calls.append("extract")

        def tweak(current_code_name):
            self.assertEqual(current_code_name, "jaunty")
            calls.append("tweak")

        def upgrade(code_name, operation_id, allow_third_party=False,
                    debug=False, mode=None):
            self.assertEqual(operation_id, 100)
            self.assertEqual(code_name, "karmic")
            self.assertTrue(allow_third_party)
            self.assertFalse(debug)
            self.assertIdentical(mode, None)
            calls.append("upgrade")

        def finish():
            calls.append("finish")

        self.upgrader.fetch = fetch
        self.upgrader.verify = verify
        self.upgrader.extract = extract
        self.upgrader.tweak = tweak
        self.upgrader.upgrade = upgrade
        self.upgrader.finish = finish

        self.upgrader.lsb_release_filename = self.makeFile(
            "DISTRIB_CODENAME=jaunty\n")

        message = {"type": "release-upgrade",
                   "code-name": "karmic",
                   "upgrade-tool-tarball-url": "http://some/tarball",
                   "upgrade-tool-signature-url": "http://some/sign",
                   "allow-third-party": True,
                   "operation-id": 100}

        result = self.upgrader.handle_release_upgrade(message)

        def check_result(ignored):
            self.assertEqual(calls, ["fetch", "verify", "extract", "tweak",
                                     "upgrade", "finish"])

        result.addCallback(check_result)
        return result

    def test_handle_release_upgrade_with_already_upgraded_system(self):
        """
        The L{ReleaseUpgrader.handle_release_upgrade} method reports a
        failure if the system is already running the desired release.
        """
        self.upgrader.lsb_release_filename = self.makeFile(
            "DISTRIB_CODENAME=karmic\n")

        message = {"type": "release-upgrade",
                   "code-name": "karmic",
                   "operation-id": 100}

        result = self.upgrader.handle_release_upgrade(message)

        def check_result(ignored):
            self.assertIn("INFO: Queuing message with release upgrade "
                          "failure to exchange urgently.",
                          self.logfile.getvalue())
            self.assertMessages(self.get_pending_messages(),
                                [{"type": "operation-result",
                                  "operation-id": 100,
                                  "status": FAILED,
                                  "result-text": "The system is already "
                                                 "running karmic.",
                                  "result-code": 1}])

        result.addCallback(check_result)
        return result

    def test_handle_release_upgrade_with_abort(self):
        """
        The L{ReleaseUpgrader.handle_release_upgrade} method reports a
        failure if any of the helper method errbacks.
        """
        self.upgrader.lsb_release_filename = self.makeFile(
            "DISTRIB_CODENAME=jaunty\n")

        calls = []

        def fetch(tarball_url, signature_url):
            calls.append("fetch")
            return succeed(None)

        def verify(tarball_filename, signature_filename):
            calls.append("verify")
            raise Exception("failure")

        def extract(tarball_filename):
            calls.append("extract")

        def tweak(current_code_name):
            calls.append("extract")

        def upgrade(code_name, operation_id):
            calls.append("upgrade")

        def finish():
            calls.append("finish")

        self.upgrader.fetch = fetch
        self.upgrader.verify = verify
        self.upgrader.extract = extract
        self.upgrader.tweak = tweak
        self.upgrader.upgrade = upgrade
        self.upgrader.finish = finish

        message = {"type": "release-upgrade",
                   "code-name": "karmic",
                   "operation-id": 100,
                   "upgrade-tool-tarball-url": "http://some/tarball",
                   "upgrade-tool-signature-url": "http://some/signature"}

        result = self.upgrader.handle_release_upgrade(message)

        def check_result(ignored):
            self.assertIn("INFO: Queuing message with release upgrade "
                          "failure to exchange urgently.",
                          self.logfile.getvalue())
            self.assertMessages(self.get_pending_messages(),
                                [{"type": "operation-result",
                                  "operation-id": 100,
                                  "status": FAILED,
                                  "result-text": "failure",
                                  "result-code": 1}])
            self.assertEqual(calls, ["fetch", "verify"])

        result.addCallback(check_result)
        return result

    def test_handle_task(self):
        """
        The L{ReleaseUpgrader.handle_task} method invokes the correct
        handler for tasks carrying messages of type C{release-upgrade}.
        """
        self.upgrader.handle_release_upgrade = lambda message: message

        message = {"type": "release-upgrade"}

        class FakeTask(object):
            data = message

        task = FakeTask()
        self.assertEqual(self.upgrader.handle_task(task), task.data)

    def test_handle_task_with_wrong_type(self):
        """
        The L{ReleaseUpgrader.handle_task} method doesn't take any action
        if the message carried by task is not known.
        """
        message = {"type": "foo"}

        class FakeTask(object):
            data = message

        self.assertEqual(self.upgrader.handle_task(FakeTask()), None)

    def test_main(self):
        """
        The L{main} function creates a new session if the process is not
        running in its own process group.
        """
        self.mocker.order()

        run_task_handler = self.mocker.replace("landscape.package.taskhandler"
                                               ".run_task_handler",
                                               passthrough=False)
        getpgrp = self.mocker.replace("os.getpgrp")
        self.expect(getpgrp()).result(os.getpid() + 1)
        setsid = self.mocker.replace("os.setsid")
        setsid()

        run_task_handler(ReleaseUpgrader, ["ARGS"])
        self.mocker.result("RESULT")

        self.mocker.replay()

        self.assertEqual(main(["ARGS"]), "RESULT")



from datetime import datetime


def to_timestamp(date, epoch=datetime.utcfromtimestamp(0)):
    """Convert a C{datetime} to an C{int}-based timetamp."""
    delta = date - epoch
    return (delta.days * 60 * 60 * 24) + delta.seconds



import logging
import os

from twisted.internet.utils import getProcessOutput

from landscape.package.store import PackageStore
from landscape.package.reporter import find_reporter_command
from landscape.monitor.plugin import MonitorPlugin


class PackageMonitor(MonitorPlugin):

    run_interval = 1800
    scope = "package"

    def __init__(self, package_store_filename=None):
        super(PackageMonitor, self).__init__()
        if package_store_filename:
            self._package_store = PackageStore(package_store_filename)
        else:
            self._package_store = None
        self._reporter_command = find_reporter_command()

    def register(self, registry):
        self.config = registry.config
        self.run_interval = self.config.package_monitor_interval
        if self.config.clones and self.config.is_clone:
            # Run clones a bit more frequently in order to catch up
            self.run_interval = 60  # 300
        super(PackageMonitor, self).register(registry)

        if not self._package_store:
            filename = os.path.join(registry.config.data_path,
                                    "package/database")
            self._package_store = PackageStore(filename)

        registry.register_message("package-ids",
                                  self._enqueue_message_as_reporter_task)
        registry.reactor.call_on("server-uuid-changed",
                                 self._server_uuid_changed)
        self.call_on_accepted("packages", self.spawn_reporter)
        self.run()

    def _enqueue_message_as_reporter_task(self, message):
        self._package_store.add_task("reporter", message)
        self.spawn_reporter()

    def run(self):
        result = self.registry.broker.get_accepted_message_types()
        result.addCallback(self._got_message_types)
        return result

    def _got_message_types(self, message_types):
        if "packages" in message_types:
            self.spawn_reporter()

    def _run_fake_reporter(self, args):
        """Run a fake-reporter in-process."""

        class FakeFacade(object):
            """
            A fake facade to workaround the issue that the AptFacade
            essentially allows only once instance per process.
            """

            def get_arch(self):
                arch = os.uname()[-1]
                result = {"pentium": "i386",
                          "i86pc": "i386",
                          "x86_64": "amd64"}.get(arch)
                if result:
                    arch = result
                elif (arch[0] == "i" and arch.endswith("86")):
                    arch = "i386"
                return arch

        if getattr(self, "_fake_reporter", None) is None:

            from landscape.package.reporter import (
                FakeReporter, PackageReporterConfiguration)
            from landscape.package.store import FakePackageStore
            package_facade = FakeFacade()
            package_config = PackageReporterConfiguration()
            package_config.load(args + ["-d", self.config.data_path,
                                        "-l", self.config.log_dir])
            package_store = FakePackageStore(package_config.store_filename)
            self._fake_reporter = FakeReporter(package_store, package_facade,
                                               self.registry.broker,
                                               package_config)
            self._fake_reporter.global_store_filename = os.path.join(
                self.config.master_data_path, "package", "database")
            self._fake_reporter_running = False

        if self._fake_reporter_running:
            from twisted.internet.defer import succeed
            return succeed(None)

        self._fake_reporter_running = True
        result = self._fake_reporter.run()

        def done(passthrough):
            self._fake_reporter_running = False
            return passthrough

        return result.addBoth(done)

    def spawn_reporter(self):
        args = ["--quiet"]
        if self.config.config:
            args.extend(["-c", self.config.config])
        env = os.environ.copy()

        if self.config.clones > 0:
            if self.config.is_clone:
                return self._run_fake_reporter(args)
            else:
                env["FAKE_GLOBAL_PACKAGE_STORE"] = "1"

        # path is set to None so that getProcessOutput does not
        # chdir to "." see bug #211373
        result = getProcessOutput(self._reporter_command,
                                  args=args, env=env,
                                  errortoo=1,
                                  path=None)
        result.addCallback(self._got_reporter_output)
        return result

    def _got_reporter_output(self, output):
        if output:
            logging.warning("Package reporter output:\n%s" % output)

    def _reset(self):
        """
        Remove all tasks *except* the resynchronize task.  This is
        because if we clear all tasks, then add the resynchronize,
        it's possible that the reporter may be running a task at this
        time and when it finishes, it will unknowningly remove the
        resynchronize task because sqlite resets its serial primary
        keys when you delete an entire table.  This problem is avoided
        by adding the task first and removing them all *except* the
        resynchronize task and not causing sqlite to reset the serial
        key.
        """
        task = self._package_store.add_task("reporter",
                                            {"type": "resynchronize"})
        self._package_store.clear_tasks(except_tasks=(task,))

    def _server_uuid_changed(self, old_uuid, new_uuid):
        """Called when the broker sends a server-uuid-changed event.

        The package hash=>id map is server-specific, so when we change
        servers, we should reset this map.
        """
        # If the old_uuid is None, it means we're just starting to
        # communicate with a server that knows how to report its UUID,
        # so we don't clear our knowledge.
        if old_uuid is not None:
            self._package_store.clear_hash_ids()



"""Bridge client side plugins to the C{MessageExchange}.

The C{BrokerServer} provides C{BrokerClient}s with a mechanism to send messages
to the server and, likewise, triggers those plugins to take action when a
exchange is impending or resynchronisaton is required.

Each C{BrokerClient} has to be registered using the
L{BrokerServer.register_client} method, after which two way communications is
possible between the C{BrokerServer} and the C{BrokerClient}.

Resynchronisation Sequence
==========================

See L{landscape.broker.exchange} sequence diagram for origin of the
"resynchronize-clients" event.

Diagram::


 1. [event 1]               --->  BrokerServer        : Event
                                                      : "resynchronize-clients"

 2. [event 2]               <---  BrokerServer        : Broadcast event
                                                      : "resynchronize"

 3. [optional: various L{BrowserClientPlugin}s respond
               to the "resynchronize" event to reset
               themselves and start report afresh.]
     (See: L{landscape.monitor.packagemonitor.PackageMonitor}
           L{landscape.monitor.plugin.MonitorPlugin}
           L{landscape.manager.keystonetoken.KeystoneToken}
           L{landscape.monitor.activeprocessinfo.ActiveProcessInfo} )


 4. [event 1]               ---> MessageExchange      : Event
    (NOTE, this is the same event as step 1           : "resynchronize-clients"
     it is handled by both BrokerServer and
     MessageExchange.
     See MessageExchange._resynchronize )

 5. MessageExchange         ---> MessageExchange      : Schedule urgent
                                                      : exchange

"""

import logging

from twisted.internet.defer import Deferred

from landscape.lib.twisted_util import gather_results
from landscape.amp import remote
from landscape.manager.manager import FAILED


def event(method):
    """Turns a L{BrokerServer} method into an event broadcaster.

    When the decorated method is called, an event is fired on all connected
    clients. The event will have the same name as the method being called,
    except that any underscore in the method name will be replaced with a dash.
    """
    event_type = method.__name__.replace("_", "-")

    def broadcast_event(self, *args, **kwargs):
        fired = []
        for client in self.get_clients():
            fired.append(client.fire_event(event_type, *args, **kwargs))
        return gather_results(fired)

    return broadcast_event


class BrokerServer(object):
    """
    A broker server capable of handling messages from plugins connected using
    the L{BrokerProtocol}.

    @param config: The L{BrokerConfiguration} used by the broker.
    @param reactor: The L{LandscapeReactor} driving the broker's events.
    @param exchange: The L{MessageExchange} to send messages with.
    @param registration: The {RegistrationHandler}.
    @param message_store: The broker's L{MessageStore}.
    """
    name = "broker"

    def __init__(self, config, reactor, exchange, registration,
                 message_store, pinger):
        from landscape.broker.amp import get_component_registry
        self.connectors_registry = get_component_registry()
        self._config = config
        self._reactor = reactor
        self._exchanger = exchange
        self._registration = registration
        self._message_store = message_store
        self._registered_clients = {}
        self._connectors = {}
        self._pinger = pinger

        reactor.call_on("message", self.broadcast_message)
        reactor.call_on("impending-exchange", self.impending_exchange)
        reactor.call_on("message-type-acceptance-changed",
                        self.message_type_acceptance_changed)
        reactor.call_on("server-uuid-changed", self.server_uuid_changed)
        reactor.call_on("package-data-changed", self.package_data_changed)
        reactor.call_on("resynchronize-clients", self.resynchronize)

    @remote
    def ping(self):
        """Return C{True}."""
        return True

    @remote
    def get_session_id(self, scope=None):
        """Get a unique session ID to be used when sending messages.

        Anything that wants to send a message to the server via the broker is
        required to first acquire a session ID with this method. Such session
        IDs must be passed to L{send_message} whenever sending a message.

        The broker keeps track of the session IDs that it hands out and will
        drop them when a re-synchronisation event occurs.  Further messages
        sent using expired IDs will be silently discarded.

        For example each L{BrokerClientPlugin} calls this method to get a
        session ID and use it when sending messages, until the plugin gets
        notified of a re-synchronisation event and then asks for a new one.

        This eliminates issues with out-of-date messages being delivered to the
        server after a re-synchronisation request. For example when the client
        re-registers and gets a new computer ID we don't want to deliver
        messages containing references to activity IDs of the old computer
        (e.g. a message with the result of a "change-packages" activity
        delivered before re-registering). See also #328005 and #1158822.

        """
        return self._message_store.get_session_id(scope=scope)

    @remote
    def register_client(self, name):
        """Register a broker client called C{name}.

        Various broker clients interact with the broker server, such as the
        monitor for example, using the L{BrokerServerConnector} for performing
        remote method calls on the L{BrokerServer}.

        They establish connectivity with the broker by connecting and
        registering themselves, the L{BrokerServer} will in turn connect
        to them in order to be able to perform remote method calls like
        broadcasting events and messages.

        @param name: The name of the client, such a C{monitor} or C{manager}.
        """
        connector_class = self.connectors_registry.get(name)
        connector = connector_class(self._reactor, self._config)

        def register(remote_client):
            self._registered_clients[name] = remote_client
            self._connectors[remote_client] = connector

        connected = connector.connect()
        return connected.addCallback(register)

    def get_clients(self):
        """Get L{RemoteClient} instances for registered clients."""
        return self._registered_clients.values()

    def get_client(self, name):
        """Return the client with the given C{name} or C{None}."""
        return self._registered_clients.get(name)

    def get_connectors(self):
        """Get connectors for registered clients.

        @see L{RemoteLandscapeComponentCreator}.
        """
        return self._connectors.values()

    def get_connector(self, name):
        """Return the connector for the given C{name} or C{None}."""
        return self._connectors.get(self.get_client(name))

    @remote
    def send_message(self, message, session_id, urgent=False):
        """Queue C{message} for delivery to the server at the next exchange.

        @param message: The message C{dict} to send to the server.  It must
            have a C{type} key and be compatible with C{landscape.lib.bpickle}.
        @param session_id: A session ID.  You should acquire this
            with C{get_session_id} before attempting to send a message.
        @param urgent: If C{True}, exchange urgently, otherwise exchange
            during the next regularly scheduled exchange.
        @return: The message identifier created when queuing C{message}.
        """
        if isinstance(session_id, bool) and message["type"] in (
                "operation-result", "change-packages-result"):
            # XXX This means we're performing a Landscape-driven upgrade and
            # we're being invoked by a package-changer or release-upgrader
            # process that is running code which doesn't know about the
            # session_id parameter, i.e. it was written against the old
            # signature:
            #
            #     def send_message(self, message, urgent=False):
            #
            # In this case we'll just let the message in. This transitional
            # code can be dropped once we stop supporting clients without
            # session IDs (i.e. with versions < 14.01).
            urgent = True
            session_id = self.get_session_id()

        if session_id is None:
            raise RuntimeError(
                "Session ID must be set before attempting to send a message")
        if self._message_store.is_valid_session_id(session_id):
            return self._exchanger.send(message, urgent=urgent)

    @remote
    def is_message_pending(self, message_id):
        """Indicate if a message with given C{message_id} is pending."""
        return self._message_store.is_pending(message_id)

    @remote
    def stop_clients(self):
        """Tell all the clients to exit."""
        results = []
        # FIXME: check whether the client are still alive
        for client in self.get_clients():
            results.append(client.exit())
        result = gather_results(results, consume_errors=True)
        return result.addCallback(lambda ignored: None)

    @remote
    def reload_configuration(self):
        """Reload the configuration file, and stop all clients."""
        self._config.reload()
        # Now we'll kill off everything else so that they can be restarted and
        # notice configuration changes.
        return self.stop_clients()

    @remote
    def register(self):
        """Attempt to register with the Landscape server.

        @see: L{RegistrationHandler.register}
        """
        return self._registration.register()

    @remote
    def get_accepted_message_types(self):
        """Return the message types accepted by the Landscape server."""
        return self._message_store.get_accepted_types()

    @remote
    def get_server_uuid(self):
        """Return the uuid of the Landscape server we're pointing at."""
        return self._message_store.get_server_uuid()

    @remote
    def register_client_accepted_message_type(self, type):
        """Register a new message type which can be accepted by this client.

        @param type: The message type to accept.
        """
        self._exchanger.register_client_accepted_message_type(type)

    @remote
    def fire_event(self, event_type):
        """Fire an event in the broker reactor."""
        self._reactor.fire(event_type)

    @remote
    def exit(self):
        """Request a graceful exit from the broker server.

        Before this method returns, all broker clients will be notified
        of the server broker's intention of exiting, so that they have
        the chance to stop whatever they're doing in a graceful way, and
        then exit themselves.

        This method will only return a result when all plugins returned
        their own results.
        """
        clients_stopped = self.stop_clients()

        def schedule_reactor_stop(ignored):
            # Stop the reactor with a short delay to give us a chance to reply
            # to the caller when this method is invoked over AMP (typically
            # by the watchdog, see landscape.watchdog.Watchdog.request_exit).
            #
            # Note that stopping the reactor will cause the Twisted machinery
            # to invoke BrokerService.stopService, which in turn will stop the
            # exchanger/pinger and cleanly close all AMP sockets.
            self._reactor.call_later(1, lambda: self._reactor.stop())

        return clients_stopped.addBoth(schedule_reactor_stop)

    @event
    def resynchronize(self):
        """Broadcast a C{resynchronize} event to the clients."""

    @event
    def impending_exchange(self):
        """Broadcast an C{impending-exchange} event to the clients."""

    @remote
    def listen_events(self, event_types):
        """
        Return a C{Deferred} that fires when the first event occurs among the
        given ones.
        """
        deferred = Deferred()
        calls = []

        def get_handler(event_type):

            def handler():
                for call in calls:
                    self._reactor.cancel_call(call)
                deferred.callback(event_type)

            return handler

        for event_type in event_types:
            call = self._reactor.call_on(event_type, get_handler(event_type))
            calls.append(call)
        return deferred

    @event
    def broker_reconnect(self):
        """Broadcast a C{broker-reconnect} event to the clients."""

    @event
    def server_uuid_changed(self, old_uuid, new_uuid):
        """Broadcast a C{server-uuid-changed} event to the clients."""

    @event
    def message_type_acceptance_changed(self, type, accepted):
        pass

    @event
    def package_data_changed(self):
        """Fire a package-data-changed event in the reactor of each client."""

    def broadcast_message(self, message):
        """Call the C{message} method of all the registered plugins.

        @see: L{register_plugin}.
        """
        results = []
        for client in self.get_clients():
            results.append(client.message(message))
        result = gather_results(results)
        return result.addCallback(self._message_delivered, message)

    def _message_delivered(self, results, message):
        """
        If the message wasn't handled, and it's an operation request (i.e. it
        has an operation-id), then respond with a failing operation result
        indicating as such.
        """
        opid = message.get("operation-id")
        if (True not in results
            and opid is not None
            and message["type"] != "resynchronize"):
            mtype = message["type"]
            logging.error("Nobody handled the %s message." % (mtype,))

            result_text = """\
Landscape client failed to handle this request (%s) because the
plugin which should handle it isn't available.  This could mean that the
plugin has been intentionally disabled, or that the client isn't running
properly, or you may be running an older version of the client that doesn't
support this feature.
""" % (mtype,)
            response = {
                "type": "operation-result",
                "status": FAILED,
                "result-text": result_text,
                "operation-id": opid}
            self._exchanger.send(response, urgent=True)

    @remote
    def stop_exchanger(self):
        """
        Stop exchaging messages with the message server.

        Eventually, it is required by the plugin that no more message exchanges
        are performed.
        For example, when a reboot process in running, the client stops
        accepting new messages so that no client action is running while the
        machine is rebooting.
        Also, some activities should be explicitly require that no more
        messages are exchanged so some level of serialization in the client
        could be achieved.
        """
        self._exchanger.stop()
        self._pinger.stop()



import pwd
import grp

from landscape.user.provider import (UserProvider, UserNotFoundError,
    GroupNotFoundError)

from landscape.user.tests.helpers import FakeUserProvider

from landscape.tests.helpers import LandscapeTest


class ProviderTest(LandscapeTest):

    def setUp(self):
        LandscapeTest.setUp(self)
        self.shadow_file = self.makeFile("""\
jdoe:$1$xFlQvTqe$cBtrNEDOIKMy/BuJoUdeG0:13348:0:99999:7:::
psmith:!:13348:0:99999:7:::
sbarnes:$1$q7sz09uw$q.A3526M/SHu8vUb.Jo1A/:13349:0:99999:7:::
""")

        self.passwd_file = self.makeFile("""\
root:x:0:0:root:/root:/bin/bash
haldaemon:x:107:116:Hardware abstraction layer,,,:/home/haldaemon:/bin/false
kevin:x:1001:65534:Kevin,101,+44123123,+44123124:/home/kevin:/bin/bash
""")

        self.group_file = self.makeFile("""\
root:x:0:
cdrom:x:24:haldaemon,kevin
kevin:x:1000:
""")

    def test_get_uid(self):
        """
        Given a username L{UserProvider.get_uid} returns the UID or
        raises a L{UserProviderError} if a match isn't found.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file)
        self.assertEqual(provider.get_uid("jdoe"), 1000)
        self.assertRaises(UserNotFoundError, provider.get_uid, "john")

    def test_get_users(self):
        """Get users should return data for all users found on the system."""
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file)
        users = provider.get_users()
        self.assertEqual(users, [{"username": "jdoe", "name": u"JD",
                                  "uid": 1000, "enabled": True,
                                  "location": None, "home-phone": None,
                                  "work-phone": None,
                                  "primary-gid": 1000}])

    def test_gecos_data(self):
        """
        Location, home phone number, and work phone number should be
        correctly parsed out of the GECOS field, and included in the
        users message.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,Everywhere,7654321,123HOME,",
                 "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file)
        users = provider.get_users()
        self.assertEqual(users, [{"username": "jdoe", "name": u"JD",
                                  "uid": 1000, "enabled": True,
                                  "location": u"Everywhere",
                                  "home-phone": u"123HOME",
                                  "work-phone": u"7654321",
                                  "primary-gid": 1000}])

    def test_four_gecos_fields(self):
        """If a GECOS field only has four fields it should still work."""
        data = [("jdoe", "x", 1000, 1000, "JD,Everywhere,7654321,123HOME",
                 "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file)
        users = provider.get_users()
        self.assertEqual(users, [{"username": "jdoe", "name": u"JD",
                                  "uid": 1000, "enabled": True,
                                  "location": u"Everywhere",
                                  "home-phone": u"123HOME",
                                  "work-phone": u"7654321",
                                  "primary-gid": 1000}])

    def test_old_school_gecos_data(self):
        """
        If C{useradd} is used to add users to a system the GECOS field
        will be written as a comment.  The client must be resilient to
        this situation.
        """
        data = [("jdoe", "x", 1000, 1000, "John Doe",
                 "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file)
        users = provider.get_users()
        self.assertEqual(users, [{"username": "jdoe", "uid": 1000,
                                  "enabled": True, "name": u"John Doe",
                                  "location": None, "home-phone": None,
                                  "work-phone": None, "primary-gid": 1000}])

    def test_weird_gecos_data(self):
        """
        If GECOS data is malformed in a way that contains less than
        four fields, read as many as are available.
        """
        data = [("jdoe", "x", 1000, 1000, "John Doe,Everywhere",
                 "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file)
        users = provider.get_users()
        self.assertEqual(users, [{"username": "jdoe", "uid": 1000,
                                  "enabled": True, "name": "John Doe",
                                  "location": "Everywhere",
                                  "home-phone": None, "work-phone": None,
                                  "primary-gid": 1000}])

    def test_no_gecos_data(self):
        """
        When no data is provided in the GECOS field we should report
        all optional fields as C{None}.
        """
        data = [("jdoe", "x", 1000, 1000, "", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file)
        users = provider.get_users()
        self.assertEqual(users, [{"username": "jdoe", "uid": 1000,
                                  "enabled": True, "name": None,
                                  "location": None, "home-phone": None,
                                  "work-phone": None,
                                  "primary-gid": 1000}])

    def test_utf8_gecos_data(self):
        """Gecos fields should be decoded from utf-8 to unicode."""
        name = u"Jos\N{LATIN SMALL LETTER E WITH ACUTE}"
        location = "F\N{LATIN SMALL LETTER I WITH DIAERESIS}nland"
        number = "N\N{LATIN SMALL LETTER AE}ver"
        gecos = "%s,%s,%s,%s," % (name.encode("utf-8"),
                                  location.encode("utf-8"),
                                  number.encode("utf-8"),
                                  number.encode("utf-8"))
        data = [("jdoe", "x", 1000, 1000, gecos, "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file)
        users = provider.get_users()
        self.assertEqual(users[0]["name"], name)
        self.assertEqual(users[0]["location"], location)
        self.assertEqual(users[0]["home-phone"], number)
        self.assertEqual(users[0]["work-phone"], number)

    def test_non_utf8_data(self):
        """
        If a GECOS field contains non-UTF8 data, it should be replaced
        with question marks.
        """
        unicode_unknown = u'\N{REPLACEMENT CHARACTER}'
        data = [("jdoe", "x", 1000, 1000, "\255,\255,\255,\255", "/home/jdoe",
                 "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file)
        users = provider.get_users()
        self.assertEqual(users[0]["name"], unicode_unknown)
        self.assertEqual(users[0]["location"], unicode_unknown)
        self.assertEqual(users[0]["home-phone"], unicode_unknown)
        self.assertEqual(users[0]["work-phone"], unicode_unknown)

    def test_get_disabled_user(self):
        """The C{enabled} field should be C{False} for disabled users."""
        data = [("psmith", "x", 1000, 1000,
                 "Peter Smith,,,,", "/home/psmith", "/bin/bash")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    locked_users=["psmith"])
        users = provider.get_users()
        self.assertEqual(users, [
                {"username": "psmith", "name": u"Peter Smith", "uid": 1000,
                 "enabled": False,
                 "location": None, "home-phone": None, "work-phone": None,
                 "primary-gid": 1000}])

    def test_real_user_data(self):
        """L{UserProvider} should work with real data."""
        provider = UserProvider()
        provider.shadow_file = None
        users = provider.get_users()
        user_0 = pwd.getpwuid(0)
        for user in users:
            if user["username"] == user_0.pw_name:
                self.assertEqual(user["uid"], 0)
                user_0_name = user_0.pw_gecos.split(",")[0].decode(
                    "utf-8", "replace")
                self.assertEqual(user["name"], user_0_name)
                break
        else:
            self.fail("The user %s (uid=0) was not found in the get_data "
                      "result." % (user_0.pw_name))

    def test_get_users_duplicate_usernames(self):
        """
        Get users should return data for all users found on the system, but it
        should exclude duplicate usernames.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh"),
                ("jdoe", "x", 1001, 1001, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file)
        users = provider.get_users()
        self.assertEqual(users, [{"username": "jdoe", "name": u"JD",
                                  "uid": 1000, "enabled": True,
                                  "location": None, "home-phone": None,
                                  "work-phone": None, "primary-gid": 1000}])

    def test_get_users_duplicate_uids(self):
        """
        Get users should return data for all users found on the system,
        including users with duplicated uids.
        """
        data = [("joe1", "x", 1000, 1000, "JD,,,,", "/home/joe1", "/bin/zsh"),
                ("joe2", "x", 1000, 1000, "JD,,,,", "/home/joe2", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file)
        users = provider.get_users()
        self.assertEqual(users, [{"username": "joe1", "name": u"JD",
                                  "uid": 1000, "enabled": True,
                                  "location": None, "home-phone": None,
                                  "work-phone": None, "primary-gid": 1000},
                                 {"username": "joe2", "name": u"JD",
                                  "uid": 1000, "enabled": True,
                                  "location": None, "home-phone": None,
                                  "work-phone": None, "primary-gid": 1000}])

    def test_user_not_in_shadow_file(self):
        """
        Given a username that doesn't exist in the shadow file, we should get a
        UserProvider error rather than a KeyError.
        raises a L{UserProviderError} if a match isn't found.
        """
        data = [("johndoe", "x", 1000, 1000,
                 "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file)
        users = provider.get_users()
        self.assertEqual(len(users), 1)
        self.assertEqual(sorted([x[0] for x in data]), ["johndoe"])

    def test_get_gid(self):
        """
        Given a username L{UserProvider.get_gid} returns the GID or
        raises a L{UserProviderError} if a match isn't found.
        """
        provider = FakeUserProvider(groups=[("jdoe", "x", 1000, [])])
        self.assertEqual(provider.get_gid("jdoe"), 1000)
        self.assertRaises(GroupNotFoundError, provider.get_gid, "john")

    def test_group_without_members(self):
        """
        L{UserProvider.get_groups} should include groups without
        members.
        """
        provider = FakeUserProvider(groups=[("jdoe", "x", 1000, [])])
        self.assertEqual(provider.get_groups(),
                         [{"name": "jdoe", "gid": 1000, "members": []}])

    def test_group_with_members(self):
        """L{UserProvider.get_groups} should include groups with members."""
        users = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        groups = [("sales", "x", 50, ["jdoe"])]
        provider = FakeUserProvider(users=users, shadow_file=self.shadow_file,
                                    groups=groups)
        self.assertEqual(provider.get_groups(),
                         [{"name": "sales", "gid": 50, "members": ["jdoe"]}])

    def test_group_with_unknown_members(self):
        """L{UserProvider.get_groups} should include groups with members.

        If a member's userid isn't known to the system, it shouldn't be
        returned.
        """
        users = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        groups = [("sales", "x", 50, ["jdoe", "kevin"])]
        provider = FakeUserProvider(users=users, shadow_file=self.shadow_file,
                                    groups=groups)
        self.assertEqual(provider.get_groups(),
                         [{"name": "sales", "gid": 50, "members": ["jdoe"]}])

    def test_group_with_duplicate_members(self):
        """
        L{UserProvider.get_groups} should only report groups once.
        If duplicates exist they should be removed.  The problem
        reported in bug #118799 is related to duplicates being
        reported to the server.
        """
        users = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        groups = [("sales", "x", 50, ["jdoe", "jdoe"])]
        provider = FakeUserProvider(users=users, shadow_file=self.shadow_file,
                                    groups=groups)
        self.assertEqual(provider.get_groups(),
                         [{"name": "sales", "gid": 50, "members": ["jdoe"]}])

    def test_group_with_duplicate_groupnames(self):
        """
        L{UserProvider.get_groups} should only report members once.
        If duplicates exist they should be removed.  The problem
        reported in bug #118799 is related to duplicates being
        reported to the server.
        """
        users = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        groups = [("sales", "x", 50, ["jdoe"]),
                  ("sales", "x", 51, ["jdoe"]),
                 ]
        provider = FakeUserProvider(users=users, shadow_file=self.shadow_file,
                                    groups=groups)
        self.assertEqual(provider.get_groups(),
                         [{"name": "sales", "gid": 50, "members": ["jdoe"]}])

    def test_real_group_data(self):
        """
        Assert that L{UserProvider.get_group}'s functionality
        reflects what is accessible from the Python standard C{grp}
        module.
        """
        provider = UserProvider()
        group_0 = grp.getgrgid(0)
        groups = provider.get_groups()
        for group in groups:
            if group["name"] == group_0.gr_name:
                self.assertEqual(group["gid"], 0)
                self.assertEqual(group["members"], group_0.gr_mem)
                break
        else:
            self.fail("The group %s (gid=0) was not found in the get_data "
                      "result." % (group_0.gr_name,))

    def test_get_user_data(self):
        """This tests the functionality for parsing /etc/passwd style files."""
        provider = UserProvider(passwd_file=self.passwd_file,
                                group_file=self.group_file)
        users = provider.get_user_data()
        self.assertEqual(users[0], ("root", "x", 0, 0, "root", "/root",
                                    "/bin/bash"))
        self.assertEqual(users[1], ("haldaemon", "x", 107, 116,
                                    "Hardware abstraction layer,,,",
                                    "/home/haldaemon", "/bin/false"))
        self.assertEqual(users[2], ("kevin", "x", 1001, 65534,
                                    "Kevin,101,+44123123,+44123124",
                                    "/home/kevin", "/bin/bash"))

    def test_get_users_with_many(self):
        """
        The method get_users is responsible for translating tuples of
        information from the underlying user database into dictionaries.
        """
        provider = UserProvider(passwd_file=self.passwd_file,
                                 group_file=self.group_file)
        users = provider.get_users()
        self.assertEqual(users[0], {"username": "root",
                                    "name": u"root",
                                    "uid": 0, "enabled": True,
                                    "location": None,
                                    "home-phone": None,
                                    "work-phone": None,
                                    "primary-gid": 0})
        self.assertEqual(users[1], {"username": "haldaemon",
                                    "name": u"Hardware abstraction layer",
                                    "uid": 107,
                                    "enabled": True,
                                    "location": None,
                                    "home-phone": None,
                                    "work-phone": None,
                                    "primary-gid": 116})
        self.assertEqual(users[2], {"username": "kevin",
                                    "name": u"Kevin",
                                    "uid": 1001,
                                    "enabled": True,
                                    "location": u"101",
                                    "home-phone": u"+44123124",
                                    "work-phone": u"+44123123",
                                    "primary-gid": 65534})

    def test_get_group_data(self):
        """This tests the functionality for parsing /etc/group style files."""
        provider = UserProvider(passwd_file=self.passwd_file,
                                group_file=self.group_file)
        groups = provider.get_group_data()
        self.assertEqual(groups[0], (u"root", u"x", 0, [u""]))
        self.assertEqual(groups[1], (u"cdrom", u"x", 24,
                                     [u"haldaemon", u"kevin"]))
        self.assertEqual(groups[2], (u"kevin", u"x", 1000, [u""]))

    def test_get_groups(self):
        """
        The method get_groups is responsible for translating tuples of data
        from the underlying userdatabase into dictionaries.
        """
        provider = UserProvider(passwd_file=self.passwd_file,
                                group_file=self.group_file)
        groups = provider.get_groups()
        self.assertEqual(groups[0], {"name": u"root",
                                     "gid": 0,
                                     "members": []})
        self.assertEqual(groups[1], {"name": u"cdrom",
                                     "gid": 24,
                                     "members": [u"kevin", u"haldaemon"]})
        self.assertEqual(groups[2], {"name": u"kevin",
                                     "gid": 1000,
                                     "members": []})

    def test_get_users_incorrect_passwd_file(self):
        """
        This tests the functionality for parsing /etc/passwd style files.

        Incorrectly formatted lines according to passwd(5) should be ignored
        during processing.
        """
        passwd_file = self.makeFile("""\
root:x:0:0:root:/root:/bin/bash
broken
haldaemon:x:107:Hardware abstraction layer,,,:/home/haldaemon:/bin/false
kevin:x:1001:65534:Kevin,101,+44123123,+44123124:/home/kevin:/bin/bash
+::::::
broken2
""")

        provider = UserProvider(passwd_file=passwd_file,
                                group_file=self.group_file)
        users = provider.get_users()
        self.assertEqual(users[0], {"username": "root",
                                    "name": u"root",
                                    "uid": 0, "enabled": True,
                                    "location": None,
                                    "home-phone": None,
                                    "work-phone": None,
                                    "primary-gid": 0})
        self.assertEqual(users[1], {"username": "kevin",
                                    "name": u"Kevin",
                                    "uid": 1001,
                                    "enabled": True,
                                    "location": u"101",
                                    "home-phone": u"+44123124",
                                    "work-phone": u"+44123123",
                                    "primary-gid": 65534})
        log = ("WARNING: passwd file %s is incorrectly formatted: line 2." %
               passwd_file)
        self.assertIn(log, self.logfile.getvalue())
        log2 = ("WARNING: passwd file %s is incorrectly formatted: line 3." %
               passwd_file)
        self.assertIn(log2, self.logfile.getvalue())
        log3 = ("WARNING: passwd file %s is incorrectly formatted: line 6." %
               passwd_file)
        self.assertIn(log3, self.logfile.getvalue())

    def test_get_users_nis_line(self):
        """
        This tests the functionality for parsing /etc/passwd style files.

        We should ignore the specific pattern for NIS user-extensions in passwd
        files.
        """
        passwd_file = self.makeFile("""\
root:x:0:0:root:/root:/bin/bash
kevin:x:1001:65534:Kevin,101,+44123123,+44123124:/home/kevin:/bin/bash
+jkakar::::::
-radix::::::
+::::::
""")

        provider = UserProvider(passwd_file=passwd_file,
                                group_file=self.group_file)
        users = provider.get_users()
        self.assertTrue(len(users), 2)
        self.assertEqual(users[0], {"username": "root",
                                    "name": u"root",
                                    "uid": 0, "enabled": True,
                                    "location": None,
                                    "home-phone": None,
                                    "work-phone": None,
                                    "primary-gid": 0})
        self.assertEqual(users[1], {"username": "kevin",
                                    "name": u"Kevin",
                                    "uid": 1001,
                                    "enabled": True,
                                    "location": u"101",
                                    "home-phone": u"+44123124",
                                    "work-phone": u"+44123123",
                                    "primary-gid": 65534})
        log = ("WARNING: passwd file %s is incorrectly formatted" %
               passwd_file)
        self.assertTrue(log not in self.logfile.getvalue())

    def test_get_groups_incorrect_groups_file(self):
        """
        This tests the functionality for parsing /etc/group style files.

        Incorrectly formatted lines according to group(5) should be ignored
        during processing.
        """
        group_file = self.makeFile("""\
root:x:0:
cdrom:x:24:
kevin:x:kevin:
""")
        provider = UserProvider(passwd_file=self.passwd_file,
                                group_file=group_file)
        groups = provider.get_groups()
        self.assertEqual(groups[0], {"name": u"root", "gid": 0,
                                     "members": []})
        self.assertEqual(groups[1], {"name": u"cdrom", "gid": 24,
                                     "members": []})
        log = ("WARNING: group file %s is incorrectly "
               "formatted: line 3." % group_file)
        self.assertIn(log, self.logfile.getvalue())

    def test_get_groups_nis_line(self):
        """
        This tests the functionality for parsing /etc/group style files.

        We should ignore the specific pattern for NIS user-extensions in
        group files.
        """
        group_file = self.makeFile("""\
root:x:0:
cdrom:x:24:
+jkakar:::
-radix:::
+:::
""")
        provider = UserProvider(passwd_file=self.passwd_file,
                                group_file=group_file)
        groups = provider.get_groups()
        self.assertEqual(groups[0], {"name": u"root", "gid": 0,
                                     "members": []})
        log = ("WARNING: group file %s is incorrectly formatted" % group_file)
        self.assertTrue(log not in self.logfile.getvalue())



import os
from landscape.tests.helpers import LandscapeTest

from landscape.manager.keystonetoken import KeystoneToken
from landscape.tests.helpers import ManagerHelper, FakePersist


class KeystoneTokenTest(LandscapeTest):

    helpers = [ManagerHelper]

    def setUp(self):
        super(KeystoneTokenTest, self).setUp()
        self.keystone_file = os.path.join(self.makeDir(), "keystone.conf")
        self.plugin = KeystoneToken(self.keystone_file)

    def test_get_keystone_token_nonexistent(self):
        """
        The plugin provides no data when the keystone configuration file
        doesn't exist.
        """
        self.assertIs(None, self.plugin.get_data())

    def test_get_keystone_token_empty(self):
        """
        The plugin provides no data when the keystone configuration file is
        empty.
        """
        self.log_helper.ignore_errors("KeystoneToken: No admin_token found .*")
        self.makeFile(path=self.keystone_file, content="")
        self.assertIs(None, self.plugin.get_data())

    def test_get_keystone_token_no_admin_token(self):
        """
        The plugin provides no data when the keystone configuration doesn't
        have an admin_token field.
        """
        self.log_helper.ignore_errors("KeystoneToken: No admin_token found .*")
        self.makeFile(path=self.keystone_file, content="[DEFAULT]")
        self.assertIs(None, self.plugin.get_data())

    def test_get_keystone_token(self):
        """
        Finally! Some data is actually there!
        """
        self.makeFile(
            path=self.keystone_file,
            content="[DEFAULT]\nadmin_token = foobar")
        self.assertEqual("foobar", self.plugin.get_data())

    def test_get_keystone_token_non_utf8(self):
        """
        The data can be arbitrary bytes.
        """
        content = "[DEFAULT]\nadmin_token = \xff"
        self.makeFile(
            path=self.keystone_file,
            content=content)
        self.assertEqual("\xff", self.plugin.get_data())

    def test_get_message(self):
        """
        L{KeystoneToken.get_message} only returns a message when the keystone
        token has changed.
        """
        self.makeFile(
            path=self.keystone_file,
            content="[DEFAULT]\nadmin_token = foobar")
        self.plugin.register(self.manager)
        message = self.plugin.get_message()
        self.assertEqual(
            {'type': 'keystone-token', 'data': 'foobar'},
            message)
        message = self.plugin.get_message()
        self.assertIs(None, message)

    def test_flush_persists_data_to_disk(self):
        """
        The plugin's C{flush} method is called every C{flush_interval} and
        creates the perists file.
        """
        flush_interval = self.config.flush_interval
        persist_filename = os.path.join(self.config.data_path,
                                        "keystone.bpickle")

        self.assertFalse(os.path.exists(persist_filename))
        self.manager.add(self.plugin)
        self.reactor.advance(flush_interval)
        self.assertTrue(os.path.exists(persist_filename))

    def test_resynchronize_message_calls_reset_method(self):
        """
        If the reactor fires a "resynchronize", with 'openstack' scope, the
        C{_reset} method on the keystone plugin object is called.
        """
        self.manager.add(self.plugin)
        self.plugin._persist = FakePersist()
        openstack_scope = ["openstack"]
        self.reactor.fire("resynchronize", openstack_scope)
        self.assertTrue(self.plugin._persist.called)

    def test_resynchronize_gets_new_session_id(self):
        """
        If L{KeystoneToken} reacts to a "resynchronize" event it should get a
        new session id as part of the process.
        """
        self.manager.add(self.plugin)
        session_id = self.plugin._session_id
        self.plugin._persist = FakePersist()
        self.plugin.client.broker.message_store.drop_session_ids()
        self.reactor.fire("resynchronize")
        self.assertNotEqual(session_id, self.plugin._session_id)

    def test_resynchronize_with_global_scope(self):
        """
        If the reactor fires a "resynchronize", with global scope, we act as if
        it had 'openstack' scope.
        """
        self.manager.add(self.plugin)
        self.plugin._persist = FakePersist()
        self.reactor.fire("resynchronize")
        self.assertTrue(self.plugin._persist.called)

    def test_do_not_resynchronize_with_other_scope(self):
        """
        If the reactor fires a "resynchronize", with an irrelevant scope, we do
        nothing.
        """
        self.manager.add(self.plugin)
        self.plugin._persist = FakePersist()
        disk_scope = ["disk"]
        self.reactor.fire("resynchronize", disk_scope)
        self.assertFalse(self.plugin._persist.called)

    def test_send_message_with_no_data(self):
        """
        If the plugin could not extract the C{admin_token} from the Keystone
        config file, upon exchange, C{None} is returned.
        """
        self.makeFile(path=self.keystone_file,
                      content="[DEFAULT]\nadmin_token =")
        self.manager.add(self.plugin)

        def check(result):
            self.assertIs(None, result)

        return self.plugin.exchange().addCallback(check)



from landscape.sysinfo.sysinfo import SysInfoPluginRegistry
from landscape.sysinfo.landscapelink import LandscapeLink
from landscape.tests.helpers import LandscapeTest


class LandscapeLinkTest(LandscapeTest):

    def setUp(self):
        super(LandscapeLinkTest, self).setUp()
        self.landscape_link = LandscapeLink()
        self.sysinfo = SysInfoPluginRegistry()
        self.sysinfo.add(self.landscape_link)

    def test_run_returns_succeeded_deferred(self):
        self.assertIs(None, self.successResultOf(self.landscape_link.run()))

    def test_run_adds_footnote(self):
        self.landscape_link.run()
        self.assertEqual(
            self.sysinfo.get_footnotes(),
            ["Graph this data and manage this system at:\n"
             "    https://landscape.canonical.com/"])



from landscape.tests.helpers import LandscapeTest
from landscape.patch import SQLiteUpgradeManager

from landscape.upgraders import package


class TestPackageUpgraders(LandscapeTest):

    def test_package_upgrade_manager(self):
        self.assertEqual(type(package.upgrade_manager), SQLiteUpgradeManager)



import random

from landscape.manager.plugin import ManagerPlugin
from landscape.manager.manager import SUCCEEDED


class FakePackageManager(ManagerPlugin):

    run_interval = 1800
    randint = random.randint

    def register(self, registry):
        super(FakePackageManager, self).register(registry)
        self.config = registry.config

        registry.register_message("change-packages",
                                  self.handle_change_packages)
        registry.register_message("change-package-locks",
                                  self.handle_change_package_locks)
        registry.register_message("release-upgrade",
                                  self.handle_release_upgrade)

    def _handle(self, response):
        delay = self.randint(30, 300)
        self.registry.reactor.call_later(
            delay, self.manager.broker.send_message, response,
            self._session_id, urgent=True)

    def handle_change_packages(self, message):
        response = {"type": "change-packages-result",
                    "operation-id": message.get("operation-id"),
                    "result-code": 1,
                    "result-text": "OK done."}
        return self._handle(response)

    def handle_change_package_locks(self, message):
        response = {"type": "operation-result",
                    "operation-id": message.get("operation-id"),
                    "status": SUCCEEDED,
                    "result-text": "Package locks successfully changed.",
                    "result-code": 0}
        return self._handle(response)

    def handle_release_upgrade(self, message):
        response = {"type": "operation-result",
                    "operation-id": message.get("operation-id"),
                    "status": SUCCEEDED,
                    "result-text": "Successful release upgrade.",
                    "result-code": 0}
        return self._handle(response)



from landscape.tests.helpers import LandscapeTest

from landscape.manager.store import ManagerStore


class ManagerStoreTest(LandscapeTest):

    def setUp(self):
        super(ManagerStoreTest, self).setUp()
        self.filename = self.makeFile()
        self.store = ManagerStore(self.filename)
        self.store.add_graph(1, u"file 1", u"user1")
        self.store.set_graph_accumulate(1, 1234, 1.0)

    def test_get_unknown_graph(self):
        graph = self.store.get_graph(1000)
        self.assertIdentical(graph, None)

    def test_get_graph(self):
        graph = self.store.get_graph(1)
        self.assertEqual(graph, (1, u"file 1", u"user1"))

    def test_get_graphs(self):
        graphs = self.store.get_graphs()
        self.assertEqual(graphs, [(1, u"file 1", u"user1")])

    def test_get_no_graphs(self):
        self.store.remove_graph(1)
        graphs = self.store.get_graphs()
        self.assertEqual(graphs, [])

    def test_add_graph(self):
        self.store.add_graph(2, u"file 2", u"user2")
        graph = self.store.get_graph(2)
        self.assertEqual(graph, (2, u"file 2", u"user2"))

    def test_add_update_graph(self):
        self.store.add_graph(1, u"file 2", u"user2")
        graph = self.store.get_graph(1)
        self.assertEqual(graph, (1, u"file 2", u"user2"))

    def test_remove_graph(self):
        self.store.remove_graph(1)
        graphs = self.store.get_graphs()
        self.assertEqual(graphs, [])

    def test_remove_unknown_graph(self):
        self.store.remove_graph(2)
        graphs = self.store.get_graphs()
        self.assertEqual(graphs, [(1, u"file 1", u"user1")])

    def test_get_accumulate_unknown_graph(self):
        accumulate = self.store.get_graph_accumulate(2)
        self.assertIdentical(accumulate, None)

    def test_set_accumulate_graph(self):
        self.store.set_graph_accumulate(2, 1234, 2.0)
        accumulate = self.store.get_graph_accumulate(2)
        self.assertEqual(accumulate, (2, 1234, 2.0))

    def test_update_accumulate_graph(self):
        self.store.set_graph_accumulate(1, 4567, 2.0)
        accumulate = self.store.get_graph_accumulate(1)
        self.assertEqual(accumulate, (1, 4567, 2.0))
